@@ -1,4 +1,4 @@
-/* SPDX-License-Identifier: MIT */
+// SPDX-License-Identifier: MIT
 
 #include "asm/section.hpp"
 
@@ -10,6 +10,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <utility>
 
 #include "helpers.hpp"
 
@@ -31,7 +32,7 @@ struct UnionStackEntry {
 struct SectionStackEntry {
 	Section *section;
 	Section *loadSection;
-	std::optional<std::string> scope; // Section's symbol scope
+	std::pair<Symbol const *, Symbol const *> labelScopes;
 	uint32_t offset;
 	int32_t loadOffset;
 	std::stack<UnionStackEntry> unionStack;
@@ -44,13 +45,15 @@ std::unordered_map<std::string, size_t> sectionMap; // Indexes into `sectionList
 uint32_t curOffset; // Offset into the current section (see sect_GetSymbolOffset)
 Section *currentSection = nullptr;
 static Section *currentLoadSection = nullptr;
-std::optional<std::string> currentLoadScope = std::nullopt;
+static std::pair<Symbol const *, Symbol const *> currentLoadLabelScopes = {nullptr, nullptr};
 int32_t loadOffset; // Offset into the LOAD section's parent (see sect_GetOutputOffset)
 
 // A quick check to see if we have an initialized section
-[[nodiscard]] static bool checksection() {
-	if (currentSection)
+[[nodiscard]]
+static bool requireSection() {
+	if (currentSection) {
 		return true;
+	}
 
 	error("Cannot output data outside of a SECTION\n");
 	return false;
@@ -58,12 +61,15 @@ int32_t loadOffset; // Offset into the LOAD section's parent (see sect_GetOutput
 
 // A quick check to see if we have an initialized section that can contain
 // this much initialized data
-[[nodiscard]] static bool checkcodesection() {
-	if (!checksection())
+[[nodiscard]]
+static bool requireCodeSection() {
+	if (!requireSection()) {
 		return false;
+	}
 
-	if (sect_HasData(currentSection->type))
+	if (sect_HasData(currentSection->type)) {
 		return true;
+	}
 
 	error(
 	    "Section '%s' cannot contain code or data (not ROM0 or ROMX)\n",
@@ -72,41 +78,18 @@ int32_t loadOffset; // Offset into the LOAD section's parent (see sect_GetOutput
 	return false;
 }
 
-[[nodiscard]] static bool checkSectionSize(Section const &sect, uint32_t size) {
-	uint32_t maxSize = sectionTypeInfo[sect.type].size;
-
-	// If the new size is reasonable, keep going
-	if (size <= maxSize)
-		return true;
-
-	error(
-	    "Section '%s' grew too big (max size = 0x%" PRIX32 " bytes, reached 0x%" PRIX32 ").\n",
-	    sect.name.c_str(),
-	    maxSize,
-	    size
-	);
-	return false;
-}
-
-// Check if the section has grown too much.
-[[nodiscard]] static bool reserveSpace(uint32_t delta_size) {
-	// This check is here to trap broken code that generates sections that are too big and to
-	// prevent the assembler from generating huge object files or trying to allocate too much
-	// memory.
-	// A check at the linking stage is still necessary.
-
-	// If the section has already overflowed, skip the check to avoid erroring out ad nauseam
-	if (currentSection->size != UINT32_MAX
-	    && !checkSectionSize(*currentSection, curOffset + loadOffset + delta_size))
-		// Mark the section as overflowed, to avoid repeating the error
-		currentSection->size = UINT32_MAX;
-
-	if (currentLoadSection && currentLoadSection->size != UINT32_MAX
-	    && !checkSectionSize(*currentLoadSection, curOffset + delta_size))
-		currentLoadSection->size = UINT32_MAX;
-
-	return currentSection->size != UINT32_MAX
-	       && (!currentLoadSection || currentLoadSection->size != UINT32_MAX);
+void sect_CheckSizes() {
+	for (Section const &sect : sectionList) {
+		if (uint32_t maxSize = sectionTypeInfo[sect.type].size; sect.size > maxSize) {
+			error(
+			    "Section '%s' grew too big (max size = 0x%" PRIX32 " bytes, reached 0x%" PRIX32
+			    ")\n",
+			    sect.name.c_str(),
+			    maxSize,
+			    sect.size
+			);
+		}
+	}
 }
 
 Section *sect_FindSectionByName(std::string const &name) {
@@ -129,33 +112,36 @@ static unsigned int mergeSectUnion(
 
 	// Unionized sections only need "compatible" constraints, and they end up with the strictest
 	// combination of both.
-	if (sect_HasData(type))
+	if (sect_HasData(type)) {
 		sectError("Cannot declare ROM sections as UNION\n");
+	}
 
-	if (org != (uint32_t)-1) {
+	if (org != UINT32_MAX) {
 		// If both are fixed, they must be the same
-		if (sect.org != (uint32_t)-1 && sect.org != org)
+		if (sect.org != UINT32_MAX && sect.org != org) {
 			sectError(
 			    "Section already declared as fixed at different address $%04" PRIx32 "\n", sect.org
 			);
-		else if (sect.align != 0 && (mask(sect.align) & (org - sect.alignOfs)))
+		} else if (sect.align != 0 && (mask(sect.align) & (org - sect.alignOfs))) {
 			sectError(
 			    "Section already declared as aligned to %u bytes (offset %" PRIu16 ")\n",
 			    1U << sect.align,
 			    sect.alignOfs
 			);
-		else
+		} else {
 			// Otherwise, just override
 			sect.org = org;
+		}
 
 	} else if (alignment != 0) {
 		// Make sure any fixed address given is compatible
-		if (sect.org != (uint32_t)-1) {
-			if ((sect.org - alignOffset) & mask(alignment))
+		if (sect.org != UINT32_MAX) {
+			if ((sect.org - alignOffset) & mask(alignment)) {
 				sectError(
 				    "Section already declared as fixed at incompatible address $%04" PRIx32 "\n",
 				    sect.org
 				);
+			}
 			// Check if alignment offsets are compatible
 		} else if ((alignOffset & mask(sect.align)) != (sect.alignOfs & mask(alignment))) {
 			sectError(
@@ -182,38 +168,41 @@ static unsigned int
 	// Fragments only need "compatible" constraints, and they end up with the strictest
 	// combination of both.
 	// The merging is however performed at the *end* of the original section!
-	if (org != (uint32_t)-1) {
+	if (org != UINT32_MAX) {
 		uint16_t curOrg = org - sect.size;
 
 		// If both are fixed, they must be the same
-		if (sect.org != (uint32_t)-1 && sect.org != curOrg)
+		if (sect.org != UINT32_MAX && sect.org != curOrg) {
 			sectError(
 			    "Section already declared as fixed at incompatible address $%04" PRIx32 "\n",
 			    sect.org
 			);
-		else if (sect.align != 0 && (mask(sect.align) & (curOrg - sect.alignOfs)))
+		} else if (sect.align != 0 && (mask(sect.align) & (curOrg - sect.alignOfs))) {
 			sectError(
 			    "Section already declared as aligned to %u bytes (offset %" PRIu16 ")\n",
 			    1U << sect.align,
 			    sect.alignOfs
 			);
-		else
+		} else {
 			// Otherwise, just override
 			sect.org = curOrg;
+		}
 
 	} else if (alignment != 0) {
 		int32_t curOfs = (alignOffset - sect.size) % (1U << alignment);
 
-		if (curOfs < 0)
+		if (curOfs < 0) {
 			curOfs += 1U << alignment;
+		}
 
 		// Make sure any fixed address given is compatible
-		if (sect.org != (uint32_t)-1) {
-			if ((sect.org - curOfs) & mask(alignment))
+		if (sect.org != UINT32_MAX) {
+			if ((sect.org - curOfs) & mask(alignment)) {
 				sectError(
 				    "Section already declared as fixed at incompatible address $%04" PRIx32 "\n",
 				    sect.org
 				);
+			}
 			// Check if alignment offsets are compatible
 		} else if ((curOfs & mask(sect.align)) != (sect.alignOfs & mask(alignment))) {
 			sectError(
@@ -243,13 +232,14 @@ static void mergeSections(
 ) {
 	unsigned int nbSectErrors = 0;
 
-	if (type != sect.type)
+	if (type != sect.type) {
 		sectError(
 		    "Section already exists but with type %s\n", sectionTypeInfo[sect.type].name.c_str()
 		);
+	}
 
 	if (sect.modifier != mod) {
-		sectError("Section already declared as %s section\n", sectionModNames[sect.modifier]);
+		sectError("Section already declared as SECTION %s\n", sectionModNames[sect.modifier]);
 	} else {
 		switch (mod) {
 		case SECTION_UNION:
@@ -261,11 +251,13 @@ static void mergeSections(
 			// Common checks
 
 			// If the section's bank is unspecified, override it
-			if (sect.bank == (uint32_t)-1)
+			if (sect.bank == UINT32_MAX) {
 				sect.bank = bank;
+			}
 			// If both specify a bank, it must be the same one
-			else if (bank != (uint32_t)-1 && sect.bank != bank)
+			else if (bank != UINT32_MAX && sect.bank != bank) {
 				sectError("Section already declared with different bank %" PRIu32 "\n", sect.bank);
+			}
 			break;
 
 		case SECTION_NORMAL:
@@ -276,13 +268,14 @@ static void mergeSections(
 		}
 	}
 
-	if (nbSectErrors)
+	if (nbSectErrors) {
 		fatalerror(
 		    "Cannot create section \"%s\" (%u error%s)\n",
 		    sect.name.c_str(),
 		    nbSectErrors,
 		    nbSectErrors == 1 ? "" : "s"
 		);
+	}
 }
 
 #undef sectError
@@ -312,9 +305,12 @@ static Section *createSection(
 	sect.align = alignment;
 	sect.alignOfs = alignOffset;
 
+	out_RegisterNode(sect.src);
+
 	// It is only needed to allocate memory for ROM sections.
-	if (sect_HasData(type))
+	if (sect_HasData(type)) {
 		sect.data.resize(sectionTypeInfo[type].size);
+	}
 
 	return &sect;
 }
@@ -333,11 +329,12 @@ static Section *getSection(
 
 	// First, validate parameters, and normalize them if applicable
 
-	if (bank != (uint32_t)-1) {
+	if (bank != UINT32_MAX) {
 		if (type != SECTTYPE_ROMX && type != SECTTYPE_VRAM && type != SECTTYPE_SRAM
-		    && type != SECTTYPE_WRAMX)
+		    && type != SECTTYPE_WRAMX) {
 			error("BANK only allowed for ROMX, WRAMX, SRAM, or VRAM sections\n");
-		else if (bank < sectionTypeInfo[type].firstBank || bank > sectionTypeInfo[type].lastBank)
+		} else if (bank < sectionTypeInfo[type].firstBank
+		           || bank > sectionTypeInfo[type].lastBank) {
 			error(
 			    "%s bank value $%04" PRIx32 " out of range ($%04" PRIx32 " to $%04" PRIx32 ")\n",
 			    sectionTypeInfo[type].name.c_str(),
@@ -345,6 +342,7 @@ static Section *getSection(
 			    sectionTypeInfo[type].firstBank,
 			    sectionTypeInfo[type].lastBank
 			);
+		}
 	} else if (nbbanks(type) == 1) {
 		// If the section type only has a single bank, implicitly force it
 		bank = sectionTypeInfo[type].firstBank;
@@ -359,8 +357,8 @@ static Section *getSection(
 		alignOffset = 0;
 	}
 
-	if (org != (uint32_t)-1) {
-		if (org < sectionTypeInfo[type].startAddr || org > endaddr(type))
+	if (org != UINT32_MAX) {
+		if (org < sectionTypeInfo[type].startAddr || org > endaddr(type)) {
 			error(
 			    "Section \"%s\"'s fixed address $%04" PRIx32 " is outside of range [$%04" PRIx16
 			    "; $%04" PRIx16 "]\n",
@@ -369,6 +367,7 @@ static Section *getSection(
 			    sectionTypeInfo[type].startAddr,
 			    endaddr(type)
 			);
+		}
 	}
 
 	if (alignment != 0) {
@@ -379,9 +378,10 @@ static Section *getSection(
 		// It doesn't make sense to have both alignment and org set
 		uint32_t mask = mask(alignment);
 
-		if (org != (uint32_t)-1) {
-			if ((org - alignOffset) & mask)
+		if (org != UINT32_MAX) {
+			if ((org - alignOffset) & mask) {
 				error("Section \"%s\"'s fixed address doesn't match its alignment\n", name.c_str());
+			}
 			alignment = 0; // Ignore it if it's satisfied
 		} else if (sectionTypeInfo[type].startAddr & mask) {
 			error(
@@ -414,25 +414,29 @@ static Section *getSection(
 
 // Set the current section
 static void changeSection() {
-	if (!currentUnionStack.empty())
+	if (!currentUnionStack.empty()) {
 		fatalerror("Cannot change the section within a UNION\n");
+	}
 
-	sym_SetCurrentSymbolScope(std::nullopt);
+	sym_ResetCurrentLabelScopes();
 }
 
 bool Section::isSizeKnown() const {
 	// SECTION UNION and SECTION FRAGMENT can still grow
-	if (modifier != SECTION_NORMAL)
+	if (modifier != SECTION_NORMAL) {
 		return false;
+	}
 
 	// The current section (or current load section if within one) is still growing
-	if (this == currentSection || this == currentLoadSection)
+	if (this == currentSection || this == currentLoadSection) {
 		return false;
+	}
 
 	// Any section on the stack is still growing
 	for (SectionStackEntry &entry : sectionStack) {
-		if (entry.section && entry.section->name == name)
+		if (entry.section && entry.section->name == name) {
 			return false;
+		}
 	}
 
 	return true;
@@ -446,12 +450,14 @@ void sect_NewSection(
     SectionSpec const &attrs,
     SectionModifier mod
 ) {
-	if (currentLoadSection)
-		fatalerror("Cannot change the section within a `LOAD` block\n");
-
 	for (SectionStackEntry &entry : sectionStack) {
-		if (entry.section && entry.section->name == name)
+		if (entry.section && entry.section->name == name) {
 			fatalerror("Section '%s' is already on the stack\n", name.c_str());
+		}
+	}
+
+	if (currentLoadSection) {
+		sect_EndLoadSection("SECTION");
 	}
 
 	Section *sect = getSection(name, type, org, attrs, mod);
@@ -475,11 +481,7 @@ void sect_SetLoadSection(
 	// Therefore, any interactions are NOT TESTED, so lift either of those restrictions at
 	// your own peril! ^^
 
-	if (!checkcodesection())
-		return;
-
-	if (currentLoadSection) {
-		error("`LOAD` blocks cannot be nested\n");
+	if (!requireCodeSection()) {
 		return;
 	}
 
@@ -488,21 +490,26 @@ void sect_SetLoadSection(
 		return;
 	}
 
-	if (mod == SECTION_FRAGMENT) {
-		error("`LOAD FRAGMENT` is not allowed\n");
-		return;
+	if (currentLoadSection) {
+		sect_EndLoadSection("LOAD");
 	}
 
 	Section *sect = getSection(name, type, org, attrs, mod);
 
-	currentLoadScope = sym_GetCurrentSymbolScope();
+	currentLoadLabelScopes = sym_GetCurrentLabelScopes();
 	changeSection();
 	loadOffset = curOffset - (mod == SECTION_UNION ? 0 : sect->size);
 	curOffset -= loadOffset;
 	currentLoadSection = sect;
 }
 
-void sect_EndLoadSection() {
+void sect_EndLoadSection(char const *cause) {
+	if (cause) {
+		warning(
+		    WARNING_UNTERMINATED_LOAD, "`LOAD` block without `ENDL` terminated by `%s`\n", cause
+		);
+	}
+
 	if (!currentLoadSection) {
 		error("Found `ENDL` outside of a `LOAD` block\n");
 		return;
@@ -512,7 +519,13 @@ void sect_EndLoadSection() {
 	curOffset += loadOffset;
 	loadOffset = 0;
 	currentLoadSection = nullptr;
-	sym_SetCurrentSymbolScope(currentLoadScope);
+	sym_SetCurrentLabelScopes(currentLoadLabelScopes);
+}
+
+void sect_CheckLoadClosed() {
+	if (currentLoadSection) {
+		warning(WARNING_UNTERMINATED_LOAD, "`LOAD` block without `ENDL` terminated by EOF\n");
+	}
 }
 
 Section *sect_GetSymbolSection() {
@@ -531,16 +544,18 @@ uint32_t sect_GetOutputOffset() {
 // Returns how many bytes need outputting for the specified alignment and offset to succeed
 uint32_t sect_GetAlignBytes(uint8_t alignment, uint16_t offset) {
 	Section *sect = sect_GetSymbolSection();
-	if (!sect)
+	if (!sect) {
 		return 0;
+	}
 
-	bool isFixed = sect->org != (uint32_t)-1;
+	bool isFixed = sect->org != UINT32_MAX;
 
 	// If the section is not aligned, no bytes are needed
 	// (fixed sections count as being maximally aligned for this purpose)
 	uint8_t curAlignment = isFixed ? 16 : sect->align;
-	if (curAlignment == 0)
+	if (curAlignment == 0) {
 		return 0;
+	}
 
 	// We need `(pcValue + curOffset + return value) % (1 << alignment) == offset`
 	uint16_t pcValue = isFixed ? sect->org : sect->alignOfs;
@@ -549,19 +564,22 @@ uint32_t sect_GetAlignBytes(uint8_t alignment, uint16_t offset) {
 }
 
 void sect_AlignPC(uint8_t alignment, uint16_t offset) {
-	if (!checksection())
+	if (!requireSection()) {
 		return;
+	}
 
 	Section *sect = sect_GetSymbolSection();
 	uint32_t alignSize = 1 << alignment; // Size of an aligned "block"
 
-	if (sect->org != (uint32_t)-1) {
-		if ((sect->org + curOffset - offset) % alignSize)
+	if (sect->org != UINT32_MAX) {
+		if ((sect->org + curOffset - offset) % alignSize) {
 			error(
 			    "Section's fixed address fails required alignment (PC = $%04" PRIx32 ")\n",
 			    sect->org + curOffset
 			);
-	} else if (sect->align != 0 && (((sect->alignOfs + curOffset) % (1u << sect->align)) - offset) % alignSize) {
+		}
+	} else if (sect->align != 0
+	           && (((sect->alignOfs + curOffset) % (1u << sect->align)) - offset) % alignSize) {
 		error(
 		    "Section's alignment fails required alignment (offset from section start = $%04" PRIx32
 		    ")\n",
@@ -583,28 +601,35 @@ void sect_AlignPC(uint8_t alignment, uint16_t offset) {
 }
 
 static void growSection(uint32_t growth) {
+	if (growth > 0 && curOffset > UINT32_MAX - growth) {
+		fatalerror("Section size would overflow internal counter\n");
+	}
 	curOffset += growth;
-	if (curOffset + loadOffset > currentSection->size)
-		currentSection->size = curOffset + loadOffset;
-	if (currentLoadSection && curOffset > currentLoadSection->size)
+	if (uint32_t outOffset = sect_GetOutputOffset(); outOffset > currentSection->size) {
+		currentSection->size = outOffset;
+	}
+	if (currentLoadSection && curOffset > currentLoadSection->size) {
 		currentLoadSection->size = curOffset;
+	}
 }
 
-static void writebyte(uint8_t byte) {
-	currentSection->data[sect_GetOutputOffset()] = byte;
+static void writeByte(uint8_t byte) {
+	if (uint32_t index = sect_GetOutputOffset(); index < currentSection->data.size()) {
+		currentSection->data[index] = byte;
+	}
 	growSection(1);
 }
 
-static void writeword(uint16_t b) {
-	writebyte(b & 0xFF);
-	writebyte(b >> 8);
+static void writeWord(uint16_t value) {
+	writeByte(value & 0xFF);
+	writeByte(value >> 8);
 }
 
-static void writelong(uint32_t b) {
-	writebyte(b & 0xFF);
-	writebyte(b >> 8);
-	writebyte(b >> 16);
-	writebyte(b >> 24);
+static void writeLong(uint32_t value) {
+	writeByte(value & 0xFF);
+	writeByte(value >> 8);
+	writeByte(value >> 16);
+	writeByte(value >> 24);
 }
 
 static void createPatch(PatchType type, Expression const &expr, uint32_t pcShift) {
@@ -633,8 +658,9 @@ static void endUnionMember() {
 	UnionStackEntry &member = currentUnionStack.top();
 	uint32_t memberSize = curOffset - member.start;
 
-	if (memberSize > member.size)
+	if (memberSize > member.size) {
 		member.size = memberSize;
+	}
 	curOffset = member.start;
 }
 
@@ -657,61 +683,75 @@ void sect_EndUnion() {
 }
 
 void sect_CheckUnionClosed() {
-	if (!currentUnionStack.empty())
+	if (!currentUnionStack.empty()) {
 		error("Unterminated UNION construct\n");
+	}
 }
 
-// Output an absolute byte
-void sect_AbsByte(uint8_t b) {
-	if (!checkcodesection())
-		return;
-	if (!reserveSpace(1))
+// Output a constant byte
+void sect_ConstByte(uint8_t byte) {
+	if (!requireCodeSection()) {
 		return;
+	}
 
-	writebyte(b);
+	writeByte(byte);
 }
 
-void sect_AbsByteGroup(uint8_t const *s, size_t length) {
-	if (!checkcodesection())
-		return;
-	if (!reserveSpace(length))
+// Output a string's character units as bytes
+void sect_ByteString(std::vector<int32_t> const &string) {
+	if (!requireCodeSection()) {
 		return;
+	}
 
-	while (length--)
-		writebyte(*s++);
+	for (int32_t unit : string) {
+		if (!checkNBit(unit, 8, "All character units")) {
+			break;
+		}
+	}
+
+	for (int32_t unit : string) {
+		writeByte(static_cast<uint8_t>(unit));
+	}
 }
 
-void sect_AbsWordGroup(uint8_t const *s, size_t length) {
-	if (!checkcodesection())
-		return;
-	if (!reserveSpace(length * 2))
+// Output a string's character units as words
+void sect_WordString(std::vector<int32_t> const &string) {
+	if (!requireCodeSection()) {
 		return;
+	}
 
-	while (length--)
-		writeword(*s++);
+	for (int32_t unit : string) {
+		if (!checkNBit(unit, 16, "All character units")) {
+			break;
+		}
+	}
+
+	for (int32_t unit : string) {
+		writeWord(static_cast<uint16_t>(unit));
+	}
 }
 
-void sect_AbsLongGroup(uint8_t const *s, size_t length) {
-	if (!checkcodesection())
-		return;
-	if (!reserveSpace(length * 4))
+// Output a string's character units as longs
+void sect_LongString(std::vector<int32_t> const &string) {
+	if (!requireCodeSection()) {
 		return;
+	}
 
-	while (length--)
-		writelong(*s++);
+	for (int32_t unit : string) {
+		writeLong(static_cast<uint32_t>(unit));
+	}
 }
 
 // Skip this many bytes
 void sect_Skip(uint32_t skip, bool ds) {
-	if (!checksection())
-		return;
-	if (!reserveSpace(skip))
+	if (!requireSection()) {
 		return;
+	}
 
 	if (!sect_HasData(currentSection->type)) {
 		growSection(skip);
 	} else {
-		if (!ds)
+		if (!ds) {
 			warning(
 			    WARNING_EMPTY_DATA_DIRECTIVE,
 			    "%s directive without data in ROM\n",
@@ -719,112 +759,104 @@ void sect_Skip(uint32_t skip, bool ds) {
 			    : (skip == 2) ? "DW"
 			                  : "DB"
 			);
+		}
 		// We know we're in a code SECTION
-		while (skip--)
-			writebyte(fillByte);
+		while (skip--) {
+			writeByte(fillByte);
+		}
 	}
 }
 
-// Output a relocatable byte. Checking will be done to see if it
-// is an absolute value in disguise.
+// Output a byte that can be relocatable or constant
 void sect_RelByte(Expression &expr, uint32_t pcShift) {
-	if (!checkcodesection())
-		return;
-	if (!reserveSpace(1))
+	if (!requireCodeSection()) {
 		return;
+	}
 
 	if (!expr.isKnown()) {
 		createPatch(PATCHTYPE_BYTE, expr, pcShift);
-		writebyte(0);
+		writeByte(0);
 	} else {
-		writebyte(expr.value());
+		writeByte(expr.value());
 	}
 }
 
-// Output several copies of a relocatable byte. Checking will be done to see if
-// it is an absolute value in disguise.
+// Output several bytes that can be relocatable or constant
 void sect_RelBytes(uint32_t n, std::vector<Expression> &exprs) {
-	if (!checkcodesection())
-		return;
-	if (!reserveSpace(n))
+	if (!requireCodeSection()) {
 		return;
+	}
 
 	for (uint32_t i = 0; i < n; i++) {
 		Expression &expr = exprs[i % exprs.size()];
 
 		if (!expr.isKnown()) {
 			createPatch(PATCHTYPE_BYTE, expr, i);
-			writebyte(0);
+			writeByte(0);
 		} else {
-			writebyte(expr.value());
+			writeByte(expr.value());
 		}
 	}
 }
 
-// Output a relocatable word. Checking will be done to see if
-// it's an absolute value in disguise.
+// Output a word that can be relocatable or constant
 void sect_RelWord(Expression &expr, uint32_t pcShift) {
-	if (!checkcodesection())
-		return;
-	if (!reserveSpace(2))
+	if (!requireCodeSection()) {
 		return;
+	}
 
 	if (!expr.isKnown()) {
 		createPatch(PATCHTYPE_WORD, expr, pcShift);
-		writeword(0);
+		writeWord(0);
 	} else {
-		writeword(expr.value());
+		writeWord(expr.value());
 	}
 }
 
-// Output a relocatable longword. Checking will be done to see if
-// is an absolute value in disguise.
+// Output a long that can be relocatable or constant
 void sect_RelLong(Expression &expr, uint32_t pcShift) {
-	if (!checkcodesection())
-		return;
-	if (!reserveSpace(2))
+	if (!requireCodeSection()) {
 		return;
+	}
 
 	if (!expr.isKnown()) {
 		createPatch(PATCHTYPE_LONG, expr, pcShift);
-		writelong(0);
+		writeLong(0);
 	} else {
-		writelong(expr.value());
+		writeLong(expr.value());
 	}
 }
 
-// Output a PC-relative relocatable byte. Checking will be done to see if it
-// is an absolute value in disguise.
+// Output a PC-relative byte that can be relocatable or constant
 void sect_PCRelByte(Expression &expr, uint32_t pcShift) {
-	if (!checkcodesection())
+	if (!requireCodeSection()) {
 		return;
-	if (!reserveSpace(1))
-		return;
-	Symbol const *pc = sym_GetPC();
+	}
 
-	if (!expr.isDiffConstant(pc)) {
+	if (Symbol const *pc = sym_GetPC(); !expr.isDiffConstant(pc)) {
 		createPatch(PATCHTYPE_JR, expr, pcShift);
-		writebyte(0);
+		writeByte(0);
 	} else {
 		Symbol const *sym = expr.symbolOf();
 		// The offset wraps (jump from ROM to HRAM, for example)
 		int16_t offset;
 
 		// Offset is relative to the byte *after* the operand
-		if (sym == pc)
+		if (sym == pc) {
 			offset = -2; // PC as operand to `jr` is lower than reference PC by 2
-		else
+		} else {
 			offset = sym->getValue() - (pc->getValue() + 1);
+		}
 
 		if (offset < -128 || offset > 127) {
 			error(
-			    "jr target must be between -128 and 127 bytes away, not %" PRId16
-			    "; use jp instead\n",
+			    "JR target must be between -128 and 127 bytes away, not %" PRId16
+			    "; use JP instead\n",
 			    offset
 			);
-			writebyte(0);
+			writeByte(0);
 		} else {
-			writebyte(offset);
+			writeByte(offset);
 		}
 	}
 }
@@ -835,16 +867,19 @@ void sect_BinaryFile(std::string const &name, int32_t startPos) {
 		error("Start position cannot be negative (%" PRId32 ")\n", startPos);
 		startPos = 0;
 	}
-	if (!checkcodesection())
+	if (!requireCodeSection()) {
 		return;
+	}
 
 	FILE *file = nullptr;
-	if (std::optional<std::string> fullPath = fstk_FindFile(name); fullPath)
+	if (std::optional<std::string> fullPath = fstk_FindFile(name); fullPath) {
 		file = fopen(fullPath->c_str(), "rb");
+	}
 	if (!file) {
 		if (generatedMissingIncludes) {
-			if (verbose)
+			if (verbose) {
 				printf("Aborting (-MG) on INCBIN file '%s' (%s)\n", name.c_str(), strerror(errno));
+			}
 			failedOnMissingInclude = true;
 		} else {
 			error("Error opening INCBIN file '%s': %s\n", name.c_str(), strerror(errno));
@@ -853,38 +888,37 @@ void sect_BinaryFile(std::string const &name, int32_t startPos) {
 	}
 	Defer closeFile{[&] { fclose(file); }};
 
-	int32_t fsize = -1;
-
 	if (fseek(file, 0, SEEK_END) != -1) {
-		fsize = ftell(file);
-
-		if (startPos > fsize) {
-			error("Specified start position is greater than length of file\n");
+		if (startPos > ftell(file)) {
+			error("Specified start position is greater than length of file '%s'\n", name.c_str());
 			return;
 		}
-
+		// The file is seekable; skip to the specified start position
 		fseek(file, startPos, SEEK_SET);
-
-		if (!reserveSpace(fsize - startPos))
-			return;
 	} else {
-		if (errno != ESPIPE)
+		if (errno != ESPIPE) {
 			error(
 			    "Error determining size of INCBIN file '%s': %s\n", name.c_str(), strerror(errno)
 			);
-		// The file isn't seekable, so we'll just skip bytes
-		while (startPos--)
-			(void)fgetc(file);
+		}
+		// The file isn't seekable, so we'll just skip bytes one at a time
+		while (startPos--) {
+			if (fgetc(file) == EOF) {
+				error(
+				    "Specified start position is greater than length of file '%s'\n", name.c_str()
+				);
+				return;
+			}
+		}
 	}
 
 	for (int byte; (byte = fgetc(file)) != EOF;) {
-		if (fsize == -1)
-			growSection(1);
-		writebyte(byte);
+		writeByte(byte);
 	}
 
-	if (ferror(file))
+	if (ferror(file)) {
 		error("Error reading INCBIN file '%s': %s\n", name.c_str(), strerror(errno));
+	}
 }
 
 // Output a slice of a binary file
@@ -893,26 +927,26 @@ void sect_BinaryFileSlice(std::string const &name, int32_t startPos, int32_t len
 		error("Start position cannot be negative (%" PRId32 ")\n", startPos);
 		startPos = 0;
 	}
-
 	if (length < 0) {
 		error("Number of bytes to read cannot be negative (%" PRId32 ")\n", length);
 		length = 0;
 	}
-
-	if (!checkcodesection())
+	if (!requireCodeSection()) {
 		return;
-	if (length == 0) // Don't even bother with 0-byte slices
-		return;
-	if (!reserveSpace(length))
+	}
+	if (length == 0) { // Don't even bother with 0-byte slices
 		return;
+	}
 
 	FILE *file = nullptr;
-	if (std::optional<std::string> fullPath = fstk_FindFile(name); fullPath)
+	if (std::optional<std::string> fullPath = fstk_FindFile(name); fullPath) {
 		file = fopen(fullPath->c_str(), "rb");
+	}
 	if (!file) {
 		if (generatedMissingIncludes) {
-			if (verbose)
+			if (verbose) {
 				printf("Aborting (-MG) on INCBIN file '%s' (%s)\n", name.c_str(), strerror(errno));
+			}
 			failedOnMissingInclude = true;
 		} else {
 			error("Error opening INCBIN file '%s': %s\n", name.c_str(), strerror(errno));
@@ -922,44 +956,50 @@ void sect_BinaryFileSlice(std::string const &name, int32_t startPos, int32_t len
 	Defer closeFile{[&] { fclose(file); }};
 
 	if (fseek(file, 0, SEEK_END) != -1) {
-		int32_t fsize = ftell(file);
-
-		if (startPos > fsize) {
-			error("Specified start position is greater than length of file\n");
+		if (int32_t fsize = ftell(file); startPos > fsize) {
+			error("Specified start position is greater than length of file '%s'\n", name.c_str());
 			return;
-		}
-
-		if ((startPos + length) > fsize) {
+		} else if (startPos + length > fsize) {
 			error(
-			    "Specified range in INCBIN is out of bounds (%" PRIu32 " + %" PRIu32 " > %" PRIu32
-			    ")\n",
+			    "Specified range in INCBIN file '%s' is out of bounds (%" PRIu32 " + %" PRIu32
+			    " > %" PRIu32 ")\n",
+			    name.c_str(),
 			    startPos,
 			    length,
 			    fsize
 			);
 			return;
 		}
-
+		// The file is seekable; skip to the specified start position
 		fseek(file, startPos, SEEK_SET);
 	} else {
-		if (errno != ESPIPE)
+		if (errno != ESPIPE) {
 			error(
 			    "Error determining size of INCBIN file '%s': %s\n", name.c_str(), strerror(errno)
 			);
-		// The file isn't seekable, so we'll just skip bytes
-		while (startPos--)
-			(void)fgetc(file);
+		}
+		// The file isn't seekable, so we'll just skip bytes one at a time
+		while (startPos--) {
+			if (fgetc(file) == EOF) {
+				error(
+				    "Specified start position is greater than length of file '%s'\n", name.c_str()
+				);
+				return;
+			}
+		}
 	}
 
 	while (length--) {
-		int byte = fgetc(file);
-
-		if (byte != EOF) {
-			writebyte(byte);
+		if (int byte = fgetc(file); byte != EOF) {
+			writeByte(byte);
 		} else if (ferror(file)) {
 			error("Error reading INCBIN file '%s': %s\n", name.c_str(), strerror(errno));
 		} else {
-			error("Premature end of file (%" PRId32 " bytes left to read)\n", length + 1);
+			error(
+			    "Premature end of INCBIN file '%s' (%" PRId32 " bytes left to read)\n",
+			    name.c_str(),
+			    length + 1
+			);
 		}
 	}
 }
@@ -969,7 +1009,7 @@ void sect_PushSection() {
 	sectionStack.push_front({
 	    .section = currentSection,
 	    .loadSection = currentLoadSection,
-	    .scope = sym_GetCurrentSymbolScope(),
+	    .labelScopes = sym_GetCurrentLabelScopes(),
 	    .offset = curOffset,
 	    .loadOffset = loadOffset,
 	    .unionStack = {},
@@ -978,16 +1018,18 @@ void sect_PushSection() {
 	// Reset the section scope
 	currentSection = nullptr;
 	currentLoadSection = nullptr;
-	sym_SetCurrentSymbolScope(std::nullopt);
+	sym_ResetCurrentLabelScopes();
 	std::swap(currentUnionStack, sectionStack.front().unionStack);
 }
 
 void sect_PopSection() {
-	if (sectionStack.empty())
+	if (sectionStack.empty()) {
 		fatalerror("No entries in the section stack\n");
+	}
 
-	if (currentLoadSection)
-		fatalerror("Cannot change the section within a `LOAD` block\n");
+	if (currentLoadSection) {
+		sect_EndLoadSection("POPS");
+	}
 
 	SectionStackEntry entry = sectionStack.front();
 	sectionStack.pop_front();
@@ -995,23 +1037,32 @@ void sect_PopSection() {
 	changeSection();
 	currentSection = entry.section;
 	currentLoadSection = entry.loadSection;
-	sym_SetCurrentSymbolScope(entry.scope);
+	sym_SetCurrentLabelScopes(entry.labelScopes);
 	curOffset = entry.offset;
 	loadOffset = entry.loadOffset;
 	std::swap(currentUnionStack, entry.unionStack);
 }
 
+void sect_CheckStack() {
+	if (!sectionStack.empty()) {
+		warning(WARNING_UNMATCHED_DIRECTIVE, "`PUSHS` without corresponding `POPS`\n");
+	}
+}
+
 void sect_EndSection() {
-	if (!currentSection)
+	if (!currentSection) {
 		fatalerror("Cannot end the section outside of a SECTION\n");
+	}
 
-	if (currentLoadSection)
-		fatalerror("Cannot end the section within a `LOAD` block\n");
-
-	if (!currentUnionStack.empty())
+	if (!currentUnionStack.empty()) {
 		fatalerror("Cannot end the section within a UNION\n");
+	}
+
+	if (currentLoadSection) {
+		sect_EndLoadSection("ENDSECTION");
+	}
 
 	// Reset the section scope
 	currentSection = nullptr;
-	sym_SetCurrentSymbolScope(std::nullopt);
+	sym_ResetCurrentLabelScopes();
 }
