@@ -1,10 +1,11 @@
-/* SPDX-License-Identifier: MIT */
+// SPDX-License-Identifier: MIT
 
 #include "asm/symbol.hpp"
 
 #include <inttypes.h>
 #include <stdio.h>
 #include <unordered_map>
+#include <unordered_set>
 
 #include "error.hpp"
 #include "helpers.hpp" // assume
@@ -19,11 +20,15 @@
 using namespace std::literals;
 
 std::unordered_map<std::string, Symbol> symbols;
+std::unordered_set<std::string> purgedSymbols;
 
-static std::optional<std::string> labelScope = std::nullopt; // Current section's label scope
+static Symbol const *globalScope = nullptr; // Current section's global label scope
+static Symbol const *localScope = nullptr;  // Current section's local label scope
 static Symbol *PCSymbol;
-static Symbol *_NARGSymbol;
-static Symbol *_RSSymbol;
+static Symbol *NARGSymbol;
+static Symbol *globalScopeSymbol;
+static Symbol *localScopeSymbol;
+static Symbol *RSSymbol;
 static char savedTIME[256];
 static char savedDATE[256];
 static char savedTIMESTAMP_ISO8601_LOCAL[256];
@@ -35,23 +40,38 @@ bool sym_IsPC(Symbol const *sym) {
 }
 
 void sym_ForEach(void (*callback)(Symbol &)) {
-	for (auto &it : symbols)
+	for (auto &it : symbols) {
 		callback(it.second);
+	}
 }
 
-static int32_t Callback_NARG() {
+static int32_t NARGCallback() {
 	if (MacroArgs const *macroArgs = fstk_GetCurrentMacroArgs(); macroArgs) {
 		return macroArgs->nbArgs();
 	} else {
-		error("_NARG does not make sense outside of a macro\n");
+		error("_NARG has no value outside of a macro\n");
 		return 0;
 	}
 }
 
-static int32_t CallbackPC() {
-	Section const *section = sect_GetSymbolSection();
+static std::shared_ptr<std::string> globalScopeCallback() {
+	if (!globalScope) {
+		error("\".\" has no value outside of a label scope\n");
+		return std::make_shared<std::string>("");
+	}
+	return std::make_shared<std::string>(globalScope->name);
+}
 
-	return section ? section->org + sect_GetSymbolOffset() : 0;
+static std::shared_ptr<std::string> localScopeCallback() {
+	if (!localScope) {
+		error("\"..\" has no value outside of a local label scope\n");
+		return std::make_shared<std::string>("");
+	}
+	return std::make_shared<std::string>(localScope->name);
+}
+
+static int32_t PCCallback() {
+	return sect_GetSymbolSection()->org + sect_GetSymbolOffset();
 }
 
 int32_t Symbol::getValue() const {
@@ -73,12 +93,18 @@ int32_t Symbol::getOutputValue() const {
 }
 
 ContentSpan const &Symbol::getMacro() const {
-	assume((std::holds_alternative<ContentSpan>(data)));
+	assume(std::holds_alternative<ContentSpan>(data));
 	return std::get<ContentSpan>(data);
 }
 
 std::shared_ptr<std::string> Symbol::getEqus() const {
-	assume(std::holds_alternative<std::shared_ptr<std::string>>(data));
+	assume(
+	    std::holds_alternative<std::shared_ptr<std::string>>(data)
+	    || std::holds_alternative<std::shared_ptr<std::string> (*)()>(data)
+	);
+	if (auto *callback = std::get_if<std::shared_ptr<std::string> (*)()>(&data); callback) {
+		return (*callback)();
+	}
 	return std::get<std::shared_ptr<std::string>>(data);
 }
 
@@ -93,19 +119,52 @@ static void dumpFilename(Symbol const &sym) {
 	}
 }
 
-// Update a symbol's definition filename and line
 static void updateSymbolFilename(Symbol &sym) {
 	std::shared_ptr<FileStackNode> oldSrc = std::move(sym.src);
 	sym.src = fstk_GetFileStack();
 	sym.fileLine = sym.src ? lexer_GetLineNo() : 0;
 
 	// If the old node was registered, ensure the new one is too
-	if (oldSrc && oldSrc->ID != (uint32_t)-1)
+	if (oldSrc && oldSrc->ID != UINT32_MAX) {
 		out_RegisterNode(sym.src);
+	}
+}
+
+static void alreadyDefinedError(Symbol const &sym, char const *asType) {
+	if (sym.isBuiltin && !sym_FindScopedValidSymbol(sym.name)) {
+		// `DEF()` would return false, so we should not claim the symbol is already defined
+		error("'%s' is reserved for a built-in symbol\n", sym.name.c_str());
+	} else {
+		error("'%s' already defined", sym.name.c_str());
+		if (asType) {
+			fprintf(stderr, " as %s", asType);
+		}
+		fputs(" at ", stderr);
+		dumpFilename(sym);
+	}
+}
+
+static void redefinedError(Symbol const &sym) {
+	assume(sym.isBuiltin);
+	if (!sym_FindScopedValidSymbol(sym.name)) {
+		// `DEF()` would return false, so we should not imply the symbol is already defined
+		error("'%s' is reserved for a built-in symbol\n", sym.name.c_str());
+	} else {
+		error("Built-in symbol '%s' cannot be redefined\n", sym.name.c_str());
+	}
+}
+
+static void assumeAlreadyExpanded(std::string const &symName) {
+	// Either the symbol name is `Global.local` or entirely '.'s (for scopes `.` and `..`),
+	// but cannot be unqualified `.local`
+	assume(!symName.starts_with('.') || symName.find_first_not_of('.') == symName.npos);
 }
 
-// Create a new symbol by name
 static Symbol &createSymbol(std::string const &symName) {
+	assumeAlreadyExpanded(symName);
+
+	static uint32_t nextDefIndex = 0;
+
 	Symbol &sym = symbols[symName];
 
 	sym.name = symName;
@@ -114,40 +173,84 @@ static Symbol &createSymbol(std::string const &symName) {
 	sym.section = nullptr;
 	sym.src = fstk_GetFileStack();
 	sym.fileLine = sym.src ? lexer_GetLineNo() : 0;
-	sym.ID = -1;
+	sym.ID = UINT32_MAX;
+	sym.defIndex = nextDefIndex++;
 
 	return sym;
 }
 
+static bool isAutoScoped(std::string const &symName) {
+	// `globalScope` should be global if it's defined
+	assume(!globalScope || globalScope->name.find('.') == std::string::npos);
+	// `localScope` should be qualified local if it's defined
+	assume(!localScope || localScope->name.find('.') != std::string::npos);
+
+	size_t dotPos = symName.find('.');
+
+	// If there are no dots, it's not a local label
+	if (dotPos == std::string::npos) {
+		return false;
+	}
+
+	// Label scopes `.` and `..` are the only nonlocal identifiers that start with a dot
+	if (dotPos == 0 && symName.find_first_not_of('.') == symName.npos) {
+		return false;
+	}
+
+	// Check for nothing after the dot
+	if (dotPos == symName.length() - 1) {
+		fatalerror("'%s' is a nonsensical reference to an empty local label\n", symName.c_str());
+	}
+
+	// Check for more than one dot
+	if (symName.find('.', dotPos + 1) != std::string::npos) {
+		fatalerror("'%s' is a nonsensical reference to a nested local label\n", symName.c_str());
+	}
+
+	// Check for already-qualified local label
+	if (dotPos > 0) {
+		return false;
+	}
+
+	// Check for unqualifiable local label
+	if (!globalScope) {
+		fatalerror("Unqualified local label '%s' in main scope\n", symName.c_str());
+	}
+
+	return true;
+}
+
 Symbol *sym_FindExactSymbol(std::string const &symName) {
+	assumeAlreadyExpanded(symName);
+
 	auto search = symbols.find(symName);
 	return search != symbols.end() ? &search->second : nullptr;
 }
 
 Symbol *sym_FindScopedSymbol(std::string const &symName) {
-	if (size_t dotPos = symName.find('.'); dotPos != std::string::npos) {
-		if (symName.find('.', dotPos + 1) != std::string::npos)
-			fatalerror(
-			    "'%s' is a nonsensical reference to a nested local symbol\n", symName.c_str()
-			);
-		// If auto-scoped local label, expand the name
-		if (dotPos == 0 && labelScope)
-			return sym_FindExactSymbol(*labelScope + symName);
-	}
-	return sym_FindExactSymbol(symName);
+	return sym_FindExactSymbol(isAutoScoped(symName) ? globalScope->name + symName : symName);
 }
 
 Symbol *sym_FindScopedValidSymbol(std::string const &symName) {
 	Symbol *sym = sym_FindScopedSymbol(symName);
 
-	// `@` has no value outside a section
+	// `@` has no value outside of a section
 	if (sym_IsPC(sym) && !sect_GetSymbolSection()) {
 		return nullptr;
 	}
-	// `_NARG` has no value outside a macro
-	if (sym == _NARGSymbol && !fstk_GetCurrentMacroArgs()) {
+	// `_NARG` has no value outside of a macro
+	if (sym == NARGSymbol && !fstk_GetCurrentMacroArgs()) {
 		return nullptr;
 	}
+	// `.` has no value outside of a global label scope
+	if (sym == globalScopeSymbol && !globalScope) {
+		return nullptr;
+	}
+	// `..` has no value outside of a local label scope
+	if (sym == localScopeSymbol && !localScope) {
+		return nullptr;
+	}
+
 	return sym;
 }
 
@@ -155,89 +258,113 @@ Symbol const *sym_GetPC() {
 	return PCSymbol;
 }
 
-// Purge a symbol
 void sym_Purge(std::string const &symName) {
 	Symbol *sym = sym_FindScopedValidSymbol(symName);
 
 	if (!sym) {
-		error("'%s' not defined\n", symName.c_str());
+		if (sym_IsPurgedScoped(symName)) {
+			error("'%s' was already purged\n", symName.c_str());
+		} else {
+			error("'%s' not defined\n", symName.c_str());
+		}
 	} else if (sym->isBuiltin) {
 		error("Built-in symbol '%s' cannot be purged\n", symName.c_str());
-	} else if (sym->ID != (uint32_t)-1) {
+	} else if (sym->ID != UINT32_MAX) {
 		error("Symbol \"%s\" is referenced and thus cannot be purged\n", symName.c_str());
 	} else {
-		// Do not keep a reference to the label's name after purging it
-		if (sym->name == labelScope)
-			labelScope = std::nullopt;
+		if (sym->isExported) {
+			warning(WARNING_PURGE_1, "Purging an exported symbol \"%s\"\n", symName.c_str());
+		} else if (sym->isLabel()) {
+			warning(WARNING_PURGE_2, "Purging a label \"%s\"\n", symName.c_str());
+		}
+		// Do not keep a reference to the label after purging it
+		if (sym == globalScope) {
+			globalScope = nullptr;
+		}
+		if (sym == localScope) {
+			localScope = nullptr;
+		}
+		purgedSymbols.emplace(sym->name);
 		symbols.erase(sym->name);
 	}
 }
 
-uint32_t sym_GetPCValue() {
-	Section const *sect = sect_GetSymbolSection();
+bool sym_IsPurgedExact(std::string const &symName) {
+	assumeAlreadyExpanded(symName);
 
-	if (!sect)
-		error("PC has no value outside a section\n");
-	else if (sect->org == (uint32_t)-1)
-		error("Expected constant PC but section is not fixed\n");
-	else
-		return CallbackPC();
-	return 0;
+	return purgedSymbols.find(symName) != purgedSymbols.end();
+}
+
+bool sym_IsPurgedScoped(std::string const &symName) {
+	return sym_IsPurgedExact(isAutoScoped(symName) ? globalScope->name + symName : symName);
 }
 
 int32_t sym_GetRSValue() {
-	return _RSSymbol->getOutputValue();
+	return RSSymbol->getOutputValue();
 }
 
 void sym_SetRSValue(int32_t value) {
-	updateSymbolFilename(*_RSSymbol);
-	_RSSymbol->data = value;
+	updateSymbolFilename(*RSSymbol);
+	RSSymbol->data = value;
 }
 
-// Return a constant symbol's value, assuming it's defined
 uint32_t Symbol::getConstantValue() const {
-	if (sym_IsPC(this))
-		return sym_GetPCValue();
-
-	if (isConstant())
+	if (isConstant()) {
 		return getValue();
+	}
 
-	error("\"%s\" does not have a constant value\n", name.c_str());
+	if (sym_IsPC(this)) {
+		if (!getSection()) {
+			error("PC has no value outside of a section\n");
+		} else {
+			error("PC does not have a constant value; the current section is not fixed\n");
+		}
+	} else {
+		error("\"%s\" does not have a constant value\n", name.c_str());
+	}
 	return 0;
 }
 
-// Return a constant symbol's value
 uint32_t sym_GetConstantValue(std::string const &symName) {
-	if (Symbol const *sym = sym_FindScopedSymbol(symName); sym)
+	if (Symbol const *sym = sym_FindScopedSymbol(symName); sym) {
 		return sym->getConstantValue();
+	}
 
-	error("'%s' not defined\n", symName.c_str());
+	if (sym_IsPurgedScoped(symName)) {
+		error("'%s' not defined; it was purged\n", symName.c_str());
+	} else {
+		error("'%s' not defined\n", symName.c_str());
+	}
 	return 0;
 }
 
-std::optional<std::string> const &sym_GetCurrentSymbolScope() {
-	return labelScope;
+std::pair<Symbol const *, Symbol const *> sym_GetCurrentLabelScopes() {
+	return {globalScope, localScope};
 }
 
-void sym_SetCurrentSymbolScope(std::optional<std::string> const &newScope) {
-	labelScope = newScope;
+void sym_SetCurrentLabelScopes(std::pair<Symbol const *, Symbol const *> newScopes) {
+	globalScope = std::get<0>(newScopes);
+	localScope = std::get<1>(newScopes);
+
+	// `globalScope` should be global if it's defined
+	assume(!globalScope || globalScope->name.find('.') == std::string::npos);
+	// `localScope` should be qualified local if it's defined
+	assume(!localScope || localScope->name.find('.') != std::string::npos);
+}
+
+void sym_ResetCurrentLabelScopes() {
+	globalScope = nullptr;
+	localScope = nullptr;
 }
 
-/*
- * Create a symbol that will be non-relocatable and ensure that it
- * hasn't already been defined or referenced in a context that would
- * require that it be relocatable
- * @param symName The name of the symbol to create
- * @param numeric If false, the symbol may not have been referenced earlier
- */
 static Symbol *createNonrelocSymbol(std::string const &symName, bool numeric) {
 	Symbol *sym = sym_FindExactSymbol(symName);
 
 	if (!sym) {
 		sym = &createSymbol(symName);
+		purgedSymbols.erase(sym->name);
 	} else if (sym->isDefined()) {
-		error("'%s' already defined at ", symName.c_str());
-		dumpFilename(*sym);
+		alreadyDefinedError(*sym, nullptr);
 		return nullptr; // Don't allow overriding the symbol, that'd be bad!
 	} else if (!numeric) {
 		// The symbol has already been referenced, but it's not allowed
@@ -249,12 +376,12 @@ static Symbol *createNonrelocSymbol(std::string const &symName, bool numeric) {
 	return sym;
 }
 
-// Add an equated symbol
 Symbol *sym_AddEqu(std::string const &symName, int32_t value) {
 	Symbol *sym = createNonrelocSymbol(symName, true);
 
-	if (!sym)
+	if (!sym) {
 		return nullptr;
+	}
 
 	sym->type = SYM_EQU;
 	sym->data = value;
@@ -265,15 +392,15 @@ Symbol *sym_AddEqu(std::string const &symName, int32_t value) {
 Symbol *sym_RedefEqu(std::string const &symName, int32_t value) {
 	Symbol *sym = sym_FindExactSymbol(symName);
 
-	if (!sym)
+	if (!sym) {
 		return sym_AddEqu(symName, value);
+	}
 
 	if (sym->isDefined() && sym->type != SYM_EQU) {
-		error("'%s' already defined as non-EQU at ", symName.c_str());
-		dumpFilename(*sym);
+		alreadyDefinedError(*sym, "non-EQU");
 		return nullptr;
 	} else if (sym->isBuiltin) {
-		error("Built-in symbol '%s' cannot be redefined\n", symName.c_str());
+		redefinedError(*sym);
 		return nullptr;
 	}
 
@@ -284,23 +411,12 @@ Symbol *sym_RedefEqu(std::string const &symName, int32_t value) {
 	return sym;
 }
 
-/*
- * Add a string equated symbol.
- *
- * If the desired symbol is a string it needs to be passed to this function with
- * quotes inside the string, like sym_AddString("name"s, "\"test\"), or the
- * assembler won't be able to use it with DB and similar. This is equivalent to
- * ``` name EQUS "\"test\"" ```
- *
- * If the desired symbol is a register or a number, just the terminator quotes
- * of the string are enough: sym_AddString("M_PI"s, "3.1415"). This is the same
- * as ``` M_PI EQUS "3.1415" ```
- */
 Symbol *sym_AddString(std::string const &symName, std::shared_ptr<std::string> str) {
 	Symbol *sym = createNonrelocSymbol(symName, false);
 
-	if (!sym)
+	if (!sym) {
 		return nullptr;
+	}
 
 	sym->type = SYM_EQUS;
 	sym->data = str;
@@ -310,18 +426,20 @@ Symbol *sym_AddString(std::string const &symName, std::shared_ptr<std::string> s
 Symbol *sym_RedefString(std::string const &symName, std::shared_ptr<std::string> str) {
 	Symbol *sym = sym_FindExactSymbol(symName);
 
-	if (!sym)
+	if (!sym) {
 		return sym_AddString(symName, str);
+	}
 
 	if (sym->type != SYM_EQUS) {
-		if (sym->isDefined())
-			error("'%s' already defined as non-EQUS at ", symName.c_str());
-		else
+		if (sym->isDefined()) {
+			alreadyDefinedError(*sym, "non-EQUS");
+		} else {
 			error("'%s' already referenced at ", symName.c_str());
-		dumpFilename(*sym);
+			dumpFilename(*sym);
+		}
 		return nullptr;
 	} else if (sym->isBuiltin) {
-		error("Built-in symbol '%s' cannot be redefined\n", symName.c_str());
+		redefinedError(*sym);
 		return nullptr;
 	}
 
@@ -331,19 +449,13 @@ Symbol *sym_RedefString(std::string const &symName, std::shared_ptr<std::string>
 	return sym;
 }
 
-// Alter a mutable symbol's value
 Symbol *sym_AddVar(std::string const &symName, int32_t value) {
 	Symbol *sym = sym_FindExactSymbol(symName);
 
 	if (!sym) {
 		sym = &createSymbol(symName);
 	} else if (sym->isDefined() && sym->type != SYM_VAR) {
-		error(
-		    "'%s' already defined as %s at ",
-		    symName.c_str(),
-		    sym->type == SYM_LABEL ? "label" : "constant"
-		);
-		dumpFilename(*sym);
+		alreadyDefinedError(*sym, sym->type == SYM_LABEL ? "label" : "constant");
 		return sym;
 	} else {
 		updateSymbolFilename(*sym);
@@ -355,78 +467,65 @@ Symbol *sym_AddVar(std::string const &symName, int32_t value) {
 	return sym;
 }
 
-/*
- * Add a label (aka "relocatable symbol")
- * @param symName The label's full name (so `.name` is invalid)
- * @return The created symbol
- */
 static Symbol *addLabel(std::string const &symName) {
-	assume(!symName.starts_with('.')); // The symbol name must have been expanded prior
+	assumeAlreadyExpanded(symName);
+
 	Symbol *sym = sym_FindExactSymbol(symName);
 
 	if (!sym) {
 		sym = &createSymbol(symName);
 	} else if (sym->isDefined()) {
-		error("'%s' already defined at ", symName.c_str());
-		dumpFilename(*sym);
+		alreadyDefinedError(*sym, nullptr);
 		return nullptr;
 	} else {
 		updateSymbolFilename(*sym);
 	}
 	// If the symbol already exists as a ref, just "take over" it
 	sym->type = SYM_LABEL;
-	sym->data = (int32_t)sect_GetSymbolOffset();
+	sym->data = static_cast<int32_t>(sect_GetSymbolOffset());
 	// Don't export anonymous labels
-	if (exportAll && !symName.starts_with('!'))
+	if (exportAll && !symName.starts_with('!')) {
 		sym->isExported = true;
+	}
 	sym->section = sect_GetSymbolSection();
 
-	if (sym && !sym->section)
+	if (sym && !sym->section) {
 		error("Label \"%s\" created outside of a SECTION\n", symName.c_str());
+	}
 
 	return sym;
 }
 
-// Add a local (`.name` or `Parent.name`) relocatable symbol
 Symbol *sym_AddLocalLabel(std::string const &symName) {
-	// Assuming no dots in `labelScope` if defined
-	assume(!labelScope.has_value() || labelScope->find('.') == std::string::npos);
+	// The symbol name should be local, qualified or not
+	assume(symName.find('.') != std::string::npos);
 
-	size_t dotPos = symName.find('.');
+	Symbol *sym = addLabel(isAutoScoped(symName) ? globalScope->name + symName : symName);
 
-	assume(dotPos != std::string::npos); // There should be at least one dot in `symName`
-
-	// Check for something after the dot
-	if (dotPos == symName.length() - 1) {
-		fatalerror("'%s' is a nonsensical reference to an empty local label\n", symName.c_str());
+	if (sym) {
+		localScope = sym;
 	}
-	// Check for more than one dot
-	if (symName.find('.', dotPos + 1) != std::string::npos)
-		fatalerror("'%s' is a nonsensical reference to a nested local label\n", symName.c_str());
 
-	if (dotPos == 0) {
-		if (!labelScope.has_value()) {
-			error("Unqualified local label '%s' in main scope\n", symName.c_str());
-			return nullptr;
-		}
-		return addLabel(*labelScope + symName);
-	}
-	return addLabel(symName);
+	return sym;
 }
 
-// Add a relocatable symbol
 Symbol *sym_AddLabel(std::string const &symName) {
+	// The symbol name should be global
+	assume(symName.find('.') == std::string::npos);
+
 	Symbol *sym = addLabel(symName);
 
-	// Set the symbol as the new scope
-	if (sym)
-		labelScope = sym->name;
+	if (sym) {
+		globalScope = sym;
+		// A new global scope resets the local scope
+		localScope = nullptr;
+	}
+
 	return sym;
 }
 
 static uint32_t anonLabelID = 0;
 
-// Add an anonymous label
 Symbol *sym_AddAnonLabel() {
 	if (anonLabelID == UINT32_MAX) {
 		error("Only %" PRIu32 " anonymous labels can be created!", anonLabelID);
@@ -438,12 +537,11 @@ Symbol *sym_AddAnonLabel() {
 	return addLabel(anon);
 }
 
-// Write an anonymous label's name to a buffer
 std::string sym_MakeAnonLabelName(uint32_t ofs, bool neg) {
 	uint32_t id = 0;
 
 	if (neg) {
-		if (ofs > anonLabelID)
+		if (ofs > anonLabelID) {
 			error(
 			    "Reference to anonymous label %" PRIu32 " before, when only %" PRIu32
 			    " ha%s been created so far\n",
@@ -451,19 +549,21 @@ std::string sym_MakeAnonLabelName(uint32_t ofs, bool neg) {
 			    anonLabelID,
 			    anonLabelID == 1 ? "s" : "ve"
 			);
-		else
+		} else {
 			id = anonLabelID - ofs;
+		}
 	} else {
 		ofs--; // We're referencing symbols that haven't been created yet...
-		if (ofs > UINT32_MAX - anonLabelID)
+		if (ofs > UINT32_MAX - anonLabelID) {
 			error(
 			    "Reference to anonymous label %" PRIu32 " after, when only %" PRIu32
 			    " may still be created\n",
 			    ofs + 1,
 			    UINT32_MAX - anonLabelID
 			);
-		else
+		} else {
 			id = anonLabelID + ofs;
+		}
 	}
 
 	std::string anon("!");
@@ -471,7 +571,6 @@ std::string sym_MakeAnonLabelName(uint32_t ofs, bool neg) {
 	return anon;
 }
 
-// Export a symbol
 void sym_Export(std::string const &symName) {
 	if (symName.starts_with('!')) {
 		error("Anonymous labels cannot be exported\n");
@@ -481,17 +580,18 @@ void sym_Export(std::string const &symName) {
 	Symbol *sym = sym_FindScopedSymbol(symName);
 
 	// If the symbol doesn't exist, create a ref that can be purged
-	if (!sym)
+	if (!sym) {
 		sym = sym_Ref(symName);
+	}
 	sym->isExported = true;
 }
 
-// Add a macro definition
 Symbol *sym_AddMacro(std::string const &symName, int32_t defLineNo, ContentSpan const &span) {
 	Symbol *sym = createNonrelocSymbol(symName, false);
 
-	if (!sym)
+	if (!sym) {
 		return nullptr;
+	}
 
 	sym->type = SYM_MACRO;
 	sym->data = span;
@@ -510,16 +610,7 @@ Symbol *sym_Ref(std::string const &symName) {
 	Symbol *sym = sym_FindScopedSymbol(symName);
 
 	if (!sym) {
-		if (symName.starts_with('.')) {
-			if (!labelScope.has_value())
-				fatalerror("Local label reference '%s' in main scope\n", symName.c_str());
-			std::string fullName = *labelScope + symName;
-
-			sym = &createSymbol(fullName);
-		} else {
-			sym = &createSymbol(symName);
-		}
-
+		sym = &createSymbol(isAutoScoped(symName) ? globalScope->name + symName : symName);
 		sym->type = SYM_REF;
 	}
 
@@ -535,16 +626,26 @@ void sym_SetExportAll(bool set) {
 void sym_Init(time_t now) {
 	PCSymbol = &createSymbol("@"s);
 	PCSymbol->type = SYM_LABEL;
-	PCSymbol->data = CallbackPC;
+	PCSymbol->data = PCCallback;
 	PCSymbol->isBuiltin = true;
 
-	_NARGSymbol = &createSymbol("_NARG"s);
-	_NARGSymbol->type = SYM_EQU;
-	_NARGSymbol->data = Callback_NARG;
-	_NARGSymbol->isBuiltin = true;
+	NARGSymbol = &createSymbol("_NARG"s);
+	NARGSymbol->type = SYM_EQU;
+	NARGSymbol->data = NARGCallback;
+	NARGSymbol->isBuiltin = true;
 
-	_RSSymbol = sym_AddVar("_RS"s, 0);
-	_RSSymbol->isBuiltin = true;
+	globalScopeSymbol = &createSymbol("."s);
+	globalScopeSymbol->type = SYM_EQUS;
+	globalScopeSymbol->data = globalScopeCallback;
+	globalScopeSymbol->isBuiltin = true;
+
+	localScopeSymbol = &createSymbol(".."s);
+	localScopeSymbol->type = SYM_EQUS;
+	localScopeSymbol->data = localScopeCallback;
+	localScopeSymbol->isBuiltin = true;
+
+	RSSymbol = sym_AddVar("_RS"s, 0);
+	RSSymbol->isBuiltin = true;
 
 	sym_AddString("__RGBDS_VERSION__"s, std::make_shared<std::string>(get_package_version_string()))
 	    ->isBuiltin = true;
@@ -555,7 +656,7 @@ void sym_Init(time_t now) {
 	sym_AddEqu("__RGBDS_RC__"s, PACKAGE_VERSION_RC)->isBuiltin = true;
 #endif
 
-	if (now == (time_t)-1) {
+	if (now == static_cast<time_t>(-1)) {
 		warn("Failed to determine current time");
 		// Fall back by pretending we are at the Epoch
 		now = 0;
