@@ -1,4 +1,4 @@
-/* SPDX-License-Identifier: MIT */
+// SPDX-License-Identifier: MIT
 
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -17,27 +17,26 @@
 #include "platform.hpp"
 #include "version.hpp"
 
-#define UNSPECIFIED 0x200 // Should not be in byte range
+static constexpr uint16_t UNSPECIFIED = 0x200;
+static_assert(UNSPECIFIED > 0xFF, "UNSPECIFIED should not be in byte range!");
 
-#define BANK_SIZE 0x4000
+static constexpr off_t BANK_SIZE = 0x4000;
 
 // Short options
-static char const *optstring = "Ccf:i:jk:L:l:m:n:Op:r:st:Vv";
+static char const *optstring = "Ccf:hi:jk:L:l:m:n:Op:r:st:Vv";
 
-/*
- * Equivalent long options
- * Please keep in the same order as short opts
- *
- * Also, make sure long opts don't create ambiguity:
- * A long opt's name should start with the same letter as its short opt,
- * except if it doesn't create any ambiguity (`verbose` versus `version`).
- * This is because long opt matching, even to a single char, is prioritized
- * over short opt matching
- */
+// Equivalent long options
+// Please keep in the same order as short opts.
+// Also, make sure long opts don't create ambiguity:
+// A long opt's name should start with the same letter as its short opt,
+// except if it doesn't create any ambiguity (`verbose` versus `version`).
+// This is because long opt matching, even to a single char, is prioritized
+// over short opt matching.
 static option const longopts[] = {
     {"color-only",       no_argument,       nullptr, 'C'},
     {"color-compatible", no_argument,       nullptr, 'c'},
     {"fix-spec",         required_argument, nullptr, 'f'},
+    {"help",             no_argument,       nullptr, 'h'},
     {"game-id",          required_argument, nullptr, 'i'},
     {"non-japanese",     no_argument,       nullptr, 'j'},
     {"new-licensee",     required_argument, nullptr, 'k'},
@@ -57,7 +56,7 @@ static option const longopts[] = {
 
 static void printUsage() {
 	fputs(
-	    "Usage: rgbfix [-jOsVv] [-C | -c] [-f <fix_spec>] [-i <game_id>] [-k <licensee>]\n"
+	    "Usage: rgbfix [-hjOsVv] [-C | -c] [-f <fix_spec>] [-i <game_id>] [-k <licensee>]\n"
 	    "              [-L <logo_file>] [-l <licensee_byte>] [-m <mbc_type>]\n"
 	    "              [-n <rom_version>] [-p <pad_value>] [-r <ram_size>] [-t <title_str>]\n"
 	    "              <file> ...\n"
@@ -76,15 +75,17 @@ static void printUsage() {
 
 static uint8_t nbErrors;
 
-[[gnu::format(printf, 1, 2)]] static void report(char const *fmt, ...) {
+[[gnu::format(printf, 1, 2)]]
+static void report(char const *fmt, ...) {
 	va_list ap;
 
 	va_start(ap, fmt);
 	vfprintf(stderr, fmt, ap);
 	va_end(ap);
 
-	if (nbErrors != UINT8_MAX)
+	if (nbErrors != UINT8_MAX) {
 		nbErrors++;
+	}
 }
 
 enum MbcType {
@@ -155,6 +156,7 @@ enum MbcType {
 	MBC_BAD,                // Specified MBC does not exist / syntax error
 	MBC_WRONG_FEATURES,     // MBC incompatible with specified features
 	MBC_BAD_RANGE,          // MBC number out of range
+	MBC_BAD_TPP1,           // Invalid TPP1 major or minor revision numbers
 };
 
 static void printAcceptedMBCNames() {
@@ -170,7 +172,7 @@ static void printAcceptedMBCNames() {
 	fputs("\tMBC6 ($20)\n", stderr);
 	fputs("\tMBC7+SENSOR+RUMBLE+RAM+BATTERY ($22)\n", stderr);
 	fputs("\tPOCKET_CAMERA ($FC)\n", stderr);
-	fputs("\tBANDAI_TAMA5 ($FD)\n", stderr);
+	fputs("\tBANDAI_TAMA5 ($FD) [aka TAMA5]\n", stderr);
 	fputs("\tHUC3 ($FE)\n", stderr);
 	fputs("\tHUC1+RAM+BATTERY ($FF)\n", stderr);
 
@@ -183,23 +185,23 @@ static void printAcceptedMBCNames() {
 
 static uint8_t tpp1Rev[2];
 
-/*
- * @return False on failure
- */
 static bool readMBCSlice(char const *&name, char const *expected) {
 	while (*expected) {
 		char c = *name++;
 
-		if (c == '\0') // Name too short
+		if (c == '\0') { // Name too short
 			return false;
+		}
 
-		if (c >= 'a' && c <= 'z') // Perform the comparison case-insensitive
+		if (c >= 'a' && c <= 'z') { // Perform the comparison case-insensitive
 			c = c - 'a' + 'A';
-		else if (c == '_') // Treat underscores as spaces
+		} else if (c == '_') { // Treat underscores as spaces
 			c = ' ';
+		}
 
-		if (c != *expected++)
+		if (c != *expected++) {
 			return false;
+		}
 	}
 	return true;
 }
@@ -222,11 +224,13 @@ static MbcType parseMBC(char const *name) {
 		char *endptr;
 		unsigned long mbc = strtoul(name, &endptr, base);
 
-		if (*endptr)
+		if (*endptr) {
 			return MBC_BAD;
-		if (mbc > 0xFF)
+		}
+		if (mbc > 0xFF) {
 			return MBC_BAD_RANGE;
-		return (MbcType)mbc;
+		}
+		return static_cast<MbcType>(mbc);
 
 	} else {
 		// Begin by reading the MBC type:
@@ -234,13 +238,15 @@ static MbcType parseMBC(char const *name) {
 		char const *ptr = name;
 
 		// Trim off leading whitespace
-		while (*ptr == ' ' || *ptr == '\t')
+		while (*ptr == ' ' || *ptr == '\t') {
 			ptr++;
+		}
 
 #define tryReadSlice(expected) \
 	do { \
-		if (!readMBCSlice(ptr, expected)) \
+		if (!readMBCSlice(ptr, expected)) { \
 			return MBC_BAD; \
+		} \
 	} while (0)
 
 		switch (*ptr++) {
@@ -248,8 +254,9 @@ static MbcType parseMBC(char const *name) {
 		case 'r':
 			tryReadSlice("OM");
 			// Handle optional " ONLY"
-			while (*ptr == ' ' || *ptr == '\t' || *ptr == '_')
+			while (*ptr == ' ' || *ptr == '\t' || *ptr == '_') {
 				ptr++;
+			}
 			if (*ptr == 'O' || *ptr == 'o') {
 				ptr++;
 				tryReadSlice("NLY");
@@ -324,20 +331,21 @@ static MbcType parseMBC(char const *name) {
 			case 'P': {
 				tryReadSlice("P1");
 				// Parse version
-				while (*ptr == ' ' || *ptr == '_')
+				while (*ptr == ' ' || *ptr == '_') {
 					ptr++;
+				}
 				// Major
 				char *endptr;
 				unsigned long val = strtoul(ptr, &endptr, 10);
 
 				if (endptr == ptr) {
 					report("error: Failed to parse TPP1 major revision number\n");
-					return MBC_BAD;
+					return MBC_BAD_TPP1;
 				}
 				ptr = endptr;
 				if (val != 1) {
-					report("error: RGBFIX only supports TPP1 versions 1.0\n");
-					return MBC_BAD;
+					report("error: RGBFIX only supports TPP1 version 1.0\n");
+					return MBC_BAD_TPP1;
 				}
 				tpp1Rev[0] = val;
 				tryReadSlice(".");
@@ -345,12 +353,12 @@ static MbcType parseMBC(char const *name) {
 				val = strtoul(ptr, &endptr, 10);
 				if (endptr == ptr) {
 					report("error: Failed to parse TPP1 minor revision number\n");
-					return MBC_BAD;
+					return MBC_BAD_TPP1;
 				}
 				ptr = endptr;
 				if (val > 0xFF) {
 					report("error: TPP1 minor revision number must be 8-bit\n");
-					return MBC_BAD;
+					return MBC_BAD_TPP1;
 				}
 				tpp1Rev[1] = val;
 				mbc = TPP1;
@@ -382,27 +390,33 @@ static MbcType parseMBC(char const *name) {
 
 		// Read "additional features"
 		uint8_t features = 0;
-#define RAM         (1 << 7)
-#define BATTERY     (1 << 6)
-#define TIMER       (1 << 5)
-#define RUMBLE      (1 << 4)
-#define SENSOR      (1 << 3)
-#define MULTIRUMBLE (1 << 2)
+		// clang-format off: vertically align values
+		static constexpr uint8_t RAM         = 1 << 7;
+		static constexpr uint8_t BATTERY     = 1 << 6;
+		static constexpr uint8_t TIMER       = 1 << 5;
+		static constexpr uint8_t RUMBLE      = 1 << 4;
+		static constexpr uint8_t SENSOR      = 1 << 3;
+		static constexpr uint8_t MULTIRUMBLE = 1 << 2;
+		// clang-format on
 
 		for (;;) {
 			// Trim off trailing whitespace
-			while (*ptr == ' ' || *ptr == '\t' || *ptr == '_')
+			while (*ptr == ' ' || *ptr == '\t' || *ptr == '_') {
 				ptr++;
+			}
 
 			// If done, start processing "features"
-			if (!*ptr)
+			if (!*ptr) {
 				break;
+			}
 			// We expect a '+' at this point
-			if (*ptr++ != '+')
+			if (*ptr++ != '+') {
 				return MBC_BAD;
+			}
 			// Trim off leading whitespace
-			while (*ptr == ' ' || *ptr == '\t' || *ptr == '_')
+			while (*ptr == ' ' || *ptr == '\t' || *ptr == '_') {
 				ptr++;
+			}
 
 			switch (*ptr++) {
 			case 'B': // BATTERY
@@ -427,8 +441,9 @@ static MbcType parseMBC(char const *name) {
 					break;
 				case 'A':
 				case 'a':
-					if (*ptr != 'M' && *ptr != 'm')
+					if (*ptr != 'M' && *ptr != 'm') {
 						return MBC_BAD;
+					}
 					ptr++;
 					features |= RAM;
 					break;
@@ -457,8 +472,9 @@ static MbcType parseMBC(char const *name) {
 
 		switch (mbc) {
 		case ROM:
-			if (!features)
+			if (!features) {
 				break;
+			}
 			mbc = ROM_RAM - 1;
 			static_assert(ROM_RAM + 1 == ROM_RAM_BATTERY, "Enum sanity check failed!");
 			static_assert(MBC1 + 1 == MBC1_RAM, "Enum sanity check failed!");
@@ -468,26 +484,29 @@ static MbcType parseMBC(char const *name) {
 			[[fallthrough]];
 		case MBC1:
 		case MMM01:
-			if (features == RAM)
+			if (features == RAM) {
 				mbc++;
-			else if (features == (RAM | BATTERY))
+			} else if (features == (RAM | BATTERY)) {
 				mbc += 2;
-			else if (features)
+			} else if (features) {
 				return MBC_WRONG_FEATURES;
+			}
 			break;
 
 		case MBC2:
-			if (features == BATTERY)
+			if (features == BATTERY) {
 				mbc = MBC2_BATTERY;
-			else if (features)
+			} else if (features) {
 				return MBC_WRONG_FEATURES;
+			}
 			break;
 
 		case MBC3:
 			// Handle timer, which also requires battery
 			if (features & TIMER) {
-				if (!(features & BATTERY))
+				if (!(features & BATTERY)) {
 					fprintf(stderr, "warning: MBC3+TIMER implies BATTERY\n");
+				}
 				features &= ~(TIMER | BATTERY); // Reset those bits
 				mbc = MBC3_TIMER_BATTERY;
 				// RAM is handled below
@@ -497,12 +516,13 @@ static MbcType parseMBC(char const *name) {
 			static_assert(
 			    MBC3_TIMER_BATTERY + 1 == MBC3_TIMER_RAM_BATTERY, "Enum sanity check failed!"
 			);
-			if (features == RAM)
+			if (features == RAM) {
 				mbc++;
-			else if (features == (RAM | BATTERY))
+			} else if (features == (RAM | BATTERY)) {
 				mbc += 2;
-			else if (features)
+			} else if (features) {
 				return MBC_WRONG_FEATURES;
+			}
 			break;
 
 		case MBC5:
@@ -514,12 +534,13 @@ static MbcType parseMBC(char const *name) {
 			static_assert(MBC5 + 2 == MBC5_RAM_BATTERY, "Enum sanity check failed!");
 			static_assert(MBC5_RUMBLE + 1 == MBC5_RUMBLE_RAM, "Enum sanity check failed!");
 			static_assert(MBC5_RUMBLE + 2 == MBC5_RUMBLE_RAM_BATTERY, "Enum sanity check failed!");
-			if (features == RAM)
+			if (features == RAM) {
 				mbc++;
-			else if (features == (RAM | BATTERY))
+			} else if (features == (RAM | BATTERY)) {
 				mbc += 2;
-			else if (features)
+			} else if (features) {
 				return MBC_WRONG_FEATURES;
+			}
 			break;
 
 		case MBC6:
@@ -527,47 +548,58 @@ static MbcType parseMBC(char const *name) {
 		case BANDAI_TAMA5:
 		case HUC3:
 			// No extra features accepted
-			if (features)
+			if (features) {
 				return MBC_WRONG_FEATURES;
+			}
 			break;
 
 		case MBC7_SENSOR_RUMBLE_RAM_BATTERY:
-			if (features != (SENSOR | RUMBLE | RAM | BATTERY))
+			if (features != (SENSOR | RUMBLE | RAM | BATTERY)) {
 				return MBC_WRONG_FEATURES;
+			}
 			break;
 
 		case HUC1_RAM_BATTERY:
-			if (features != (RAM | BATTERY)) // HuC1 expects RAM+BATTERY
+			if (features != (RAM | BATTERY)) { // HuC1 expects RAM+BATTERY
 				return MBC_WRONG_FEATURES;
+			}
 			break;
 
 		case TPP1:
-			if (features & RAM)
+			if (features & RAM) {
 				fprintf(
 				    stderr, "warning: TPP1 requests RAM implicitly if given a non-zero RAM size"
 				);
-			if (features & BATTERY)
+			}
+			if (features & BATTERY) {
 				mbc |= 0x08;
-			if (features & TIMER)
+			}
+			if (features & TIMER) {
 				mbc |= 0x04;
-			if (features & MULTIRUMBLE)
+			}
+			if (features & MULTIRUMBLE) {
 				mbc |= 0x03; // Also set the rumble flag
-			if (features & RUMBLE)
+			}
+			if (features & RUMBLE) {
 				mbc |= 0x01;
-			if (features & SENSOR)
+			}
+			if (features & SENSOR) {
 				return MBC_WRONG_FEATURES;
+			}
 			break;
 		}
 
 		// Trim off trailing whitespace
-		while (*ptr == ' ' || *ptr == '\t')
+		while (*ptr == ' ' || *ptr == '\t') {
 			ptr++;
+		}
 
 		// If there is still something past the whitespace, error out
-		if (*ptr)
+		if (*ptr) {
 			return MBC_BAD;
+		}
 
-		return (MbcType)mbc;
+		return static_cast<MbcType>(mbc);
 	}
 }
 
@@ -663,6 +695,7 @@ static char const *mbcName(MbcType type) {
 	case MBC_BAD:
 	case MBC_WRONG_FEATURES:
 	case MBC_BAD_RANGE:
+	case MBC_BAD_TPP1:
 		unreachable_();
 	}
 
@@ -686,6 +719,7 @@ static bool hasRAM(MbcType type) {
 	case MBC_BAD:
 	case MBC_WRONG_FEATURES:
 	case MBC_BAD_RANGE:
+	case MBC_BAD_TPP1:
 		return false;
 
 	case ROM_RAM:
@@ -737,12 +771,14 @@ static uint8_t const nintendoLogo[] = {
 };
 
 static uint8_t fixSpec = 0;
-#define FIX_LOGO         (1 << 7)
-#define TRASH_LOGO       (1 << 6)
-#define FIX_HEADER_SUM   (1 << 5)
-#define TRASH_HEADER_SUM (1 << 4)
-#define FIX_GLOBAL_SUM   (1 << 3)
-#define TRASH_GLOBAL_SUM (1 << 2)
+// clang-format off: vertically align values
+static constexpr uint8_t FIX_LOGO         = 1 << 7;
+static constexpr uint8_t TRASH_LOGO       = 1 << 6;
+static constexpr uint8_t FIX_HEADER_SUM   = 1 << 5;
+static constexpr uint8_t TRASH_HEADER_SUM = 1 << 4;
+static constexpr uint8_t FIX_GLOBAL_SUM   = 1 << 3;
+static constexpr uint8_t TRASH_GLOBAL_SUM = 1 << 2;
+// clang-format on
 
 static enum { DMG, BOTH, CGB } model = DMG; // If DMG, byte is left alone
 static char const *gameID = nullptr;
@@ -775,11 +811,13 @@ static ssize_t readBytes(int fd, uint8_t *buf, size_t len) {
 	while (len) {
 		ssize_t ret = read(fd, buf, len);
 
-		if (ret == -1 && errno != EINTR) // Return errors, unless we only were interrupted
+		if (ret == -1 && errno != EINTR) { // Return errors, unless we only were interrupted
 			return -1;
+		}
 		// EOF reached
-		if (ret == 0)
+		if (ret == 0) {
 			return total;
+		}
 		// If anything was read, accumulate it, and continue
 		if (ret != -1) {
 			total += ret;
@@ -800,43 +838,30 @@ static ssize_t writeBytes(int fd, uint8_t *buf, size_t len) {
 	while (len) {
 		ssize_t ret = write(fd, buf, len);
 
-		if (ret == -1 && errno != EINTR) // Return errors, unless we only were interrupted
+		if (ret == -1 && errno != EINTR) { // Return errors, unless we only were interrupted
 			return -1;
-		// EOF reached
-		if (ret == 0)
-			return total;
-		// If anything was read, accumulate it, and continue
+		}
+		// If anything was written, accumulate it, and continue
 		if (ret != -1) {
 			total += ret;
 			len -= ret;
+			buf += ret;
 		}
 	}
 
 	return total;
 }
 
-/*
- * @param rom0 A pointer to rom0
- * @param addr What address to check
- * @param fixedByte The fixed byte at the address
- * @param areaName Name to be displayed in the warning message
- */
 static void overwriteByte(uint8_t *rom0, uint16_t addr, uint8_t fixedByte, char const *areaName) {
 	uint8_t origByte = rom0[addr];
 
-	if (!overwriteRom && origByte != 0 && origByte != fixedByte)
+	if (!overwriteRom && origByte != 0 && origByte != fixedByte) {
 		fprintf(stderr, "warning: Overwrote a non-zero byte in the %s\n", areaName);
+	}
 
 	rom0[addr] = fixedByte;
 }
 
-/*
- * @param rom0 A pointer to rom0
- * @param startAddr What address to begin checking from
- * @param fixed The fixed bytes at the address
- * @param size How many bytes to check
- * @param areaName Name to be displayed in the warning message
- */
 static void overwriteBytes(
     uint8_t *rom0, uint16_t startAddr, uint8_t const *fixed, uint8_t size, char const *areaName
 ) {
@@ -854,18 +879,13 @@ static void overwriteBytes(
 	memcpy(&rom0[startAddr], fixed, size);
 }
 
-/*
- * @param input File descriptor to be used for reading
- * @param output File descriptor to be used for writing, may be equal to `input`
- * @param name The file's name, to be displayed for error output
- * @param fileSize The file's size if known, 0 if not.
- */
 static void processFile(int input, int output, char const *name, off_t fileSize) {
 	// Both of these should be true for seekable files, and neither otherwise
-	if (input == output)
+	if (input == output) {
 		assume(fileSize != 0);
-	else
+	} else {
 		assume(fileSize == 0);
+	}
 
 	uint8_t rom0[BANK_SIZE];
 	ssize_t rom0Len = readBytes(input, rom0, sizeof(rom0));
@@ -879,33 +899,45 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 		report(
 		    "FATAL: \"%s\" too short, expected at least %jd ($%jx) bytes, got only %jd\n",
 		    name,
-		    (intmax_t)headerSize,
-		    (intmax_t)headerSize,
-		    (intmax_t)rom0Len
+		    static_cast<intmax_t>(headerSize),
+		    static_cast<intmax_t>(headerSize),
+		    static_cast<intmax_t>(rom0Len)
 		);
 		return;
 	}
 	// Accept partial reads if the file contains at least the header
 
-	if (fixSpec & (FIX_LOGO | TRASH_LOGO))
+	if (fixSpec & (FIX_LOGO | TRASH_LOGO)) {
 		overwriteBytes(rom0, 0x0104, logo, sizeof(logo), logoFilename ? "logo" : "Nintendo logo");
+	}
 
-	if (title)
-		overwriteBytes(rom0, 0x134, (uint8_t const *)title, titleLen, "title");
+	if (title) {
+		overwriteBytes(rom0, 0x134, reinterpret_cast<uint8_t const *>(title), titleLen, "title");
+	}
 
-	if (gameID)
-		overwriteBytes(rom0, 0x13F, (uint8_t const *)gameID, gameIDLen, "manufacturer code");
-
-	if (model != DMG)
-		overwriteByte(rom0, 0x143, model == BOTH ? 0x80 : 0xC0, "CGB flag");
-
-	if (newLicensee)
+	if (gameID) {
 		overwriteBytes(
-		    rom0, 0x144, (uint8_t const *)newLicensee, newLicenseeLen, "new licensee code"
+		    rom0, 0x13F, reinterpret_cast<uint8_t const *>(gameID), gameIDLen, "manufacturer code"
 		);
+	}
 
-	if (sgb)
+	if (model != DMG) {
+		overwriteByte(rom0, 0x143, model == BOTH ? 0x80 : 0xC0, "CGB flag");
+	}
+
+	if (newLicensee) {
+		overwriteBytes(
+		    rom0,
+		    0x144,
+		    reinterpret_cast<uint8_t const *>(newLicensee),
+		    newLicenseeLen,
+		    "new licensee code"
+		);
+	}
+
+	if (sgb) {
 		overwriteByte(rom0, 0x146, 0x03, "SGB flag");
+	}
 
 	// If a valid MBC was specified...
 	if (cartridgeType < MBC_NONE) {
@@ -928,31 +960,36 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 
 		overwriteBytes(rom0, 0x150, tpp1Rev, sizeof(tpp1Rev), "TPP1 revision number");
 
-		if (ramSize != UNSPECIFIED)
+		if (ramSize != UNSPECIFIED) {
 			overwriteByte(rom0, 0x152, ramSize, "RAM size");
+		}
 
 		overwriteByte(rom0, 0x153, cartridgeType & 0xFF, "TPP1 feature flags");
 	} else {
 		// Regular mappers
 
-		if (ramSize != UNSPECIFIED)
+		if (ramSize != UNSPECIFIED) {
 			overwriteByte(rom0, 0x149, ramSize, "RAM size");
+		}
 
-		if (!japanese)
+		if (!japanese) {
 			overwriteByte(rom0, 0x14A, 0x01, "destination code");
+		}
 	}
 
-	if (oldLicensee != UNSPECIFIED)
+	if (oldLicensee != UNSPECIFIED) {
 		overwriteByte(rom0, 0x14B, oldLicensee, "old licensee code");
-	else if (sgb && rom0[0x14B] != 0x33)
+	} else if (sgb && rom0[0x14B] != 0x33) {
 		fprintf(
 		    stderr,
 		    "warning: SGB compatibility enabled, but old licensee was 0x%02x, not 0x33\n",
 		    rom0[0x14B]
 		);
+	}
 
-	if (romVersion != UNSPECIFIED)
+	if (romVersion != UNSPECIFIED) {
 		overwriteByte(rom0, 0x14C, romVersion, "mask ROM version number");
+	}
 
 	// Remain to be handled the ROM size, and header checksum.
 	// The latter depends on the former, and so will be handled after it.
@@ -1000,13 +1037,15 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 				nbBanks++;
 
 				// Update global checksum, too
-				for (uint16_t i = 0; i < bankLen; i++)
+				for (uint16_t i = 0; i < bankLen; i++) {
 					globalSum += romx[totalRomxLen + i];
+				}
 				totalRomxLen += bankLen;
 			}
 			// Stop when an incomplete bank has been read
-			if (bankLen != BANK_SIZE)
+			if (bankLen != BANK_SIZE) {
 				break;
+			}
 		}
 	}
 
@@ -1033,8 +1072,9 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 		// Alter number of banks to reflect required value
 		// x&(x-1) is zero iff x is a power of 2, or 0; we know for sure it's non-zero,
 		// so this is true (non-zero) when we don't have a power of 2
-		if (nbBanks & (nbBanks - 1))
+		if (nbBanks & (nbBanks - 1)) {
 			nbBanks = 1 << (CHAR_BIT * sizeof(nbBanks) - clz(nbBanks));
+		}
 		// Write final ROM size
 		rom0[0x148] = ctz(nbBanks / 2);
 		// Alter global checksum based on how many bytes will be added (not counting ROM0)
@@ -1045,8 +1085,9 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 	if (fixSpec & (FIX_HEADER_SUM | TRASH_HEADER_SUM)) {
 		uint8_t sum = 0;
 
-		for (uint16_t i = 0x134; i < 0x14D; i++)
+		for (uint16_t i = 0x134; i < 0x14D; i++) {
 			sum -= rom0[i] + 1;
+		}
 
 		overwriteByte(rom0, 0x14D, fixSpec & TRASH_HEADER_SUM ? ~sum : sum, "header checksum");
 	}
@@ -1054,26 +1095,33 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 	if (fixSpec & (FIX_GLOBAL_SUM | TRASH_GLOBAL_SUM)) {
 		// Computation of the global checksum does not include the checksum bytes
 		assume(rom0Len >= 0x14E);
-		for (uint16_t i = 0; i < 0x14E; i++)
+		for (uint16_t i = 0; i < 0x14E; i++) {
 			globalSum += rom0[i];
-		for (uint16_t i = 0x150; i < rom0Len; i++)
+		}
+		for (uint16_t i = 0x150; i < rom0Len; i++) {
 			globalSum += rom0[i];
+		}
 		// Pipes have already read ROMX and updated globalSum, but not regular files
 		if (input == output) {
 			for (;;) {
 				ssize_t bankLen = readBytes(input, bank, sizeof(bank));
 
-				for (uint16_t i = 0; i < bankLen; i++)
+				for (uint16_t i = 0; i < bankLen; i++) {
 					globalSum += bank[i];
-				if (bankLen != sizeof(bank))
+				}
+				if (bankLen != sizeof(bank)) {
 					break;
+				}
 			}
 		}
 
-		if (fixSpec & TRASH_GLOBAL_SUM)
+		if (fixSpec & TRASH_GLOBAL_SUM) {
 			globalSum = ~globalSum;
+		}
 
-		uint8_t bytes[2] = {(uint8_t)(globalSum >> 8), (uint8_t)(globalSum & 0xFF)};
+		uint8_t bytes[2] = {
+		    static_cast<uint8_t>(globalSum >> 8), static_cast<uint8_t>(globalSum & 0xFF)
+		};
 
 		overwriteBytes(rom0, 0x14E, bytes, sizeof(bytes), "global checksum");
 	}
@@ -1083,14 +1131,15 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 	// In case the output depends on the input, reset to the beginning of the file, and only
 	// write the header
 	if (input == output) {
-		if (lseek(output, 0, SEEK_SET) == (off_t)-1) {
+		if (lseek(output, 0, SEEK_SET) == static_cast<off_t>(-1)) {
 			report("FATAL: Failed to rewind \"%s\": %s\n", name, strerror(errno));
 			return;
 		}
 		// If modifying the file in-place, we only need to edit the header
 		// However, padding may have modified ROM0 (added padding), so don't in that case
-		if (padValue == UNSPECIFIED)
+		if (padValue == UNSPECIFIED) {
 			rom0Len = headerSize;
+		}
 	}
 	writeLen = writeBytes(output, rom0, rom0Len);
 
@@ -1100,9 +1149,9 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 	} else if (writeLen < rom0Len) {
 		report(
 		    "FATAL: Could only write %jd of \"%s\"'s %jd ROM0 bytes\n",
-		    (intmax_t)writeLen,
+		    static_cast<intmax_t>(writeLen),
 		    name,
-		    (intmax_t)rom0Len
+		    static_cast<intmax_t>(rom0Len)
 		);
 		return;
 	}
@@ -1115,10 +1164,10 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 		if (writeLen == -1) {
 			report("FATAL: Failed to write \"%s\"'s ROMX: %s\n", name, strerror(errno));
 			return;
-		} else if ((size_t)writeLen < totalRomxLen) {
+		} else if (static_cast<size_t>(writeLen) < totalRomxLen) {
 			report(
 			    "FATAL: Could only write %jd of \"%s\"'s %zu ROMX bytes\n",
-			    (intmax_t)writeLen,
+			    static_cast<intmax_t>(writeLen),
 			    name,
 			    totalRomxLen
 			);
@@ -1129,7 +1178,7 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 	// Output padding
 	if (padValue != UNSPECIFIED) {
 		if (input == output) {
-			if (lseek(output, 0, SEEK_END) == (off_t)-1) {
+			if (lseek(output, 0, SEEK_END) == static_cast<off_t>(-1)) {
 				report("FATAL: Failed to seek to end of \"%s\": %s\n", name, strerror(errno));
 				return;
 			}
@@ -1144,7 +1193,7 @@ static void processFile(int input, int output, char const *name, off_t fileSize)
 
 			// The return value is either -1, or at most `thisLen`,
 			// so it's fine to cast to `size_t`
-			if ((size_t)ret != thisLen) {
+			if (static_cast<size_t>(ret) != thisLen) {
 				report("FATAL: Failed to write \"%s\"'s padding: %s\n", name, strerror(errno));
 				break;
 			}
@@ -1157,9 +1206,9 @@ static bool processFilename(char const *name) {
 	nbErrors = 0;
 
 	if (!strcmp(name, "-")) {
+		name = "<stdin>";
 		(void)setmode(STDIN_FILENO, O_BINARY);
 		(void)setmode(STDOUT_FILENO, O_BINARY);
-		name = "<stdin>";
 		processFile(STDIN_FILENO, STDOUT_FILENO, name, 0);
 	} else {
 		// POSIX specifies that the results of O_RDWR on a FIFO are undefined.
@@ -1185,7 +1234,7 @@ static bool processFilename(char const *name) {
 				report(
 				    "FATAL: \"%s\" too short, expected at least 336 ($150) bytes, got only %jd\n",
 				    name,
-				    (intmax_t)stat.st_size
+				    static_cast<intmax_t>(stat.st_size)
 				);
 			} else {
 				processFile(input, input, name, stat.st_size);
@@ -1193,7 +1242,7 @@ static bool processFilename(char const *name) {
 		}
 	}
 
-	if (nbErrors)
+	if (nbErrors) {
 		fprintf(
 		    stderr,
 		    "Fixing \"%s\" failed with %u error%s\n",
@@ -1201,6 +1250,7 @@ static bool processFilename(char const *name) {
 		    nbErrors,
 		    nbErrors == 1 ? "" : "s"
 		);
+	}
 	return nbErrors;
 }
 
@@ -1250,16 +1300,17 @@ int main(int argc, char *argv[]) {
 				switch (*musl_optarg) {
 #define OVERRIDE_SPEC(cur, bad, curFlag, badFlag) \
 	case STR(cur)[0]: \
-		if (fixSpec & badFlag) \
+		if (fixSpec & badFlag) { \
 			fprintf(stderr, "warning: '" STR(cur) "' overriding '" STR(bad) "' in fix spec\n"); \
+		} \
 		fixSpec = (fixSpec & ~badFlag) | curFlag; \
 		break
 #define overrideSpecs(fix, fixFlag, trash, trashFlag) \
 	OVERRIDE_SPEC(fix, trash, fixFlag, trashFlag); \
 	OVERRIDE_SPEC(trash, fix, trashFlag, fixFlag)
-				overrideSpecs(l, FIX_LOGO,       L, TRASH_LOGO);
-				overrideSpecs(h, FIX_HEADER_SUM, H, TRASH_HEADER_SUM);
-				overrideSpecs(g, FIX_GLOBAL_SUM, G, TRASH_GLOBAL_SUM);
+					overrideSpecs(l, FIX_LOGO, L, TRASH_LOGO);
+					overrideSpecs(h, FIX_HEADER_SUM, H, TRASH_HEADER_SUM);
+					overrideSpecs(g, FIX_GLOBAL_SUM, G, TRASH_GLOBAL_SUM);
 #undef OVERRIDE_SPEC
 #undef overrideSpecs
 
@@ -1270,6 +1321,10 @@ int main(int argc, char *argv[]) {
 			}
 			break;
 
+		case 'h':
+			printUsage();
+			exit(0);
+
 		case 'i':
 			gameID = musl_optarg;
 			len = strlen(gameID);
@@ -1324,8 +1379,8 @@ int main(int argc, char *argv[]) {
 			} else if (cartridgeType == ROM_RAM || cartridgeType == ROM_RAM_BATTERY) {
 				fprintf(
 				    stderr,
-				    "warning: ROM+RAM / ROM+RAM+BATTERY are under-specified and poorly "
-				    "supported\n"
+				    "warning: MBC \"%s\" is under-specified and poorly supported\n",
+				    musl_optarg
 				);
 			}
 			break;
@@ -1372,27 +1427,28 @@ int main(int argc, char *argv[]) {
 			break;
 
 		default:
-			fprintf(stderr, "FATAL: unknown option '%c'\n", ch);
 			printUsage();
 			exit(1);
 		}
 	}
 
-	if ((cartridgeType & 0xFF00) == TPP1 && !japanese)
+	if ((cartridgeType & 0xFF00) == TPP1 && !japanese) {
 		fprintf(
 		    stderr,
 		    "warning: TPP1 overwrites region flag for its identification code, ignoring `-j`\n"
 		);
+	}
 
 	// Check that RAM size is correct for "standard" mappers
 	if (ramSize != UNSPECIFIED && (cartridgeType & 0xFF00) == 0) {
 		if (cartridgeType == ROM_RAM || cartridgeType == ROM_RAM_BATTERY) {
-			if (ramSize != 1)
+			if (ramSize != 1) {
 				fprintf(
 				    stderr,
 				    "warning: MBC \"%s\" should have 2 KiB of RAM (-r 1)\n",
 				    mbcName(cartridgeType)
 				);
+			}
 		} else if (hasRAM(cartridgeType)) {
 			if (!ramSize) {
 				fprintf(
@@ -1417,12 +1473,13 @@ int main(int argc, char *argv[]) {
 		}
 	}
 
-	if (sgb && oldLicensee != UNSPECIFIED && oldLicensee != 0x33)
+	if (sgb && oldLicensee != UNSPECIFIED && oldLicensee != 0x33) {
 		fprintf(
 		    stderr,
 		    "warning: SGB compatibility enabled, but old licensee is 0x%02x, not 0x33\n",
 		    oldLicensee
 		);
+	}
 
 	argv += musl_optind;
 	bool failed = nbErrors;
@@ -1433,7 +1490,8 @@ int main(int argc, char *argv[]) {
 			logoFile = fopen(logoFilename, "rb");
 		} else {
 			logoFilename = "<stdin>";
-			logoFile = fdopen(STDIN_FILENO, "rb");
+			(void)setmode(STDIN_FILENO, O_BINARY);
+			logoFile = stdin;
 		}
 		if (!logoFile) {
 			fprintf(
@@ -1472,8 +1530,9 @@ int main(int argc, char *argv[]) {
 		memcpy(logo, nintendoLogo, sizeof(nintendoLogo));
 	}
 	if (fixSpec & TRASH_LOGO) {
-		for (uint16_t i = 0; i < sizeof(logo); i++)
+		for (uint16_t i = 0; i < sizeof(logo); i++) {
 			logo[i] = 0xFF ^ logo[i];
+		}
 	}
 
 	if (!*argv) {
