@@ -1,4 +1,4 @@
-/* SPDX-License-Identifier: MIT */
+// SPDX-License-Identifier: MIT
 
 #include "gfx/main.hpp"
 
@@ -13,6 +13,7 @@
 #include <string.h>
 #include <string_view>
 #include <type_traits>
+#include <vector>
 
 #include "extern/getopt.hpp"
 #include "file.hpp"
@@ -35,15 +36,23 @@ static struct LocalOptions {
 	bool autoPalettes;
 	bool autoPalmap;
 	bool groupOutputs;
+	bool reverse;
 } localOptions;
 
 static uintmax_t nbErrors;
 
-[[noreturn]] void giveUp() {
+[[noreturn]]
+void giveUp() {
 	fprintf(stderr, "Conversion aborted after %ju error%s\n", nbErrors, nbErrors == 1 ? "" : "s");
 	exit(1);
 }
 
+void requireZeroErrors() {
+	if (nbErrors != 0) {
+		giveUp();
+	}
+}
+
 void warning(char const *fmt, ...) {
 	va_list ap;
 
@@ -63,18 +72,21 @@ void error(char const *fmt, ...) {
 	va_end(ap);
 	putc('\n', stderr);
 
-	if (nbErrors != std::numeric_limits<decltype(nbErrors)>::max())
+	if (nbErrors != std::numeric_limits<decltype(nbErrors)>::max()) {
 		nbErrors++;
+	}
 }
 
 void errorMessage(char const *msg) {
 	fprintf(stderr, "error: %s\n", msg);
 
-	if (nbErrors != std::numeric_limits<decltype(nbErrors)>::max())
+	if (nbErrors != std::numeric_limits<decltype(nbErrors)>::max()) {
 		nbErrors++;
+	}
 }
 
-[[noreturn]] void fatal(char const *fmt, ...) {
+[[noreturn]]
+void fatal(char const *fmt, ...) {
 	va_list ap;
 
 	fputs("FATAL: ", stderr);
@@ -83,8 +95,9 @@ void errorMessage(char const *msg) {
 	va_end(ap);
 	putc('\n', stderr);
 
-	if (nbErrors != std::numeric_limits<decltype(nbErrors)>::max())
+	if (nbErrors != std::numeric_limits<decltype(nbErrors)>::max()) {
 		nbErrors++;
+	}
 
 	giveUp();
 }
@@ -100,53 +113,56 @@ void Options::verbosePrint(uint8_t level, char const *fmt, ...) const {
 }
 
 // Short options
-static char const *optstring = "-Aa:b:Cc:Dd:FfhL:mN:n:Oo:Pp:Qq:r:s:Tt:U:uVvx:Z";
+static char const *optstring = "-Aa:b:Cc:d:hi:L:mN:n:Oo:Pp:Qq:r:s:Tt:U:uVvXx:YZ";
 
-/*
- * Equivalent long options
- * Please keep in the same order as short opts
- *
- * Also, make sure long opts don't create ambiguity:
- * A long opt's name should start with the same letter as its short opt,
- * except if it doesn't create any ambiguity (`verbose` versus `version`).
- * This is because long opt matching, even to a single char, is prioritized
- * over short opt matching
- */
+// Equivalent long options
+// Please keep in the same order as short opts.
+// Also, make sure long opts don't create ambiguity:
+// A long opt's name should start with the same letter as its short opt,
+// except if it doesn't create any ambiguity (`verbose` versus `version`).
+// This is because long opt matching, even to a single char, is prioritized
+// over short opt matching.
 static option const longopts[] = {
-    {"auto-attr-map",      no_argument,       nullptr, 'A' },
-    {"attr-map",           required_argument, nullptr, 'a' },
-    {"base-tiles",         required_argument, nullptr, 'b' },
-    {"color-curve",        no_argument,       nullptr, 'C' },
-    {"colors",             required_argument, nullptr, 'c' },
-    {"depth",              required_argument, nullptr, 'd' },
-    {"slice",              required_argument, nullptr, 'L' },
-    {"mirror-tiles",       no_argument,       nullptr, 'm' },
-    {"nb-tiles",           required_argument, nullptr, 'N' },
-    {"nb-palettes",        required_argument, nullptr, 'n' },
-    {"group-outputs",      no_argument,       nullptr, 'O' },
-    {"output",             required_argument, nullptr, 'o' },
-    {"auto-palette",       no_argument,       nullptr, 'P' },
-    {"palette",            required_argument, nullptr, 'p' },
-    {"auto-palette-map",   no_argument,       nullptr, 'Q' },
-    {"palette-map",        required_argument, nullptr, 'q' },
-    {"reverse",            required_argument, nullptr, 'r' },
-    {"auto-tilemap",       no_argument,       nullptr, 'T' },
-    {"tilemap",            required_argument, nullptr, 't' },
-    {"unit-size",          required_argument, nullptr, 'U' },
-    {"unique-tiles",       no_argument,       nullptr, 'u' },
-    {"version",            no_argument,       nullptr, 'V' },
-    {"verbose",            no_argument,       nullptr, 'v' },
-    {"trim-end",           required_argument, nullptr, 'x' },
-    {"columns",            no_argument,       nullptr, 'Z' },
-    {nullptr,              no_argument,       nullptr, 0   }
+    {"auto-attr-map",    no_argument,       nullptr, 'A'},
+    {"attr-map",         required_argument, nullptr, 'a'},
+    {"base-tiles",       required_argument, nullptr, 'b'},
+    {"color-curve",      no_argument,       nullptr, 'C'},
+    {"colors",           required_argument, nullptr, 'c'},
+    {"depth",            required_argument, nullptr, 'd'},
+    {"help",             no_argument,       nullptr, 'h'},
+    {"input-tileset",    required_argument, nullptr, 'i'},
+    {"slice",            required_argument, nullptr, 'L'},
+    {"mirror-tiles",     no_argument,       nullptr, 'm'},
+    {"nb-tiles",         required_argument, nullptr, 'N'},
+    {"nb-palettes",      required_argument, nullptr, 'n'},
+    {"group-outputs",    no_argument,       nullptr, 'O'},
+    {"output",           required_argument, nullptr, 'o'},
+    {"auto-palette",     no_argument,       nullptr, 'P'},
+    {"palette",          required_argument, nullptr, 'p'},
+    {"auto-palette-map", no_argument,       nullptr, 'Q'},
+    {"palette-map",      required_argument, nullptr, 'q'},
+    {"reverse",          required_argument, nullptr, 'r'},
+    {"palette-size",     required_argument, nullptr, 's'},
+    {"auto-tilemap",     no_argument,       nullptr, 'T'},
+    {"tilemap",          required_argument, nullptr, 't'},
+    {"unit-size",        required_argument, nullptr, 'U'},
+    {"unique-tiles",     no_argument,       nullptr, 'u'},
+    {"version",          no_argument,       nullptr, 'V'},
+    {"verbose",          no_argument,       nullptr, 'v'},
+    {"mirror-x",         no_argument,       nullptr, 'X'},
+    {"trim-end",         required_argument, nullptr, 'x'},
+    {"mirror-y",         no_argument,       nullptr, 'Y'},
+    {"columns",          no_argument,       nullptr, 'Z'},
+    {nullptr,            no_argument,       nullptr, 0  }
 };
 
 static void printUsage() {
 	fputs(
-	    "Usage: rgbgfx [-r stride] [-CmOuVZ] [-v [-v ...]] [-a <attr_map> | -A]\n"
-	    "       [-b <base_ids>] [-c <colors>] [-d <depth>] [-L <slice>] [-N <nb_tiles>]\n"
-	    "       [-n <nb_pals>] [-o <out_file>] [-p <pal_file> | -P] [-q <pal_map> | -Q]\n"
-	    "       [-s <nb_colors>] [-t <tile_map> | -T] [-x <nb_tiles>] <file>\n"
+	    "Usage: rgbgfx [-r stride] [-ChmOuVXYZ] [-v [-v ...]] [-a <attr_map> | -A]\n"
+	    "       [-b <base_ids>] [-c <colors>] [-d <depth>] [-i <tileset_file>]\n"
+	    "       [-L <slice>] [-N <nb_tiles>] [-n <nb_pals>] [-o <out_file>]\n"
+	    "       [-p <pal_file> | -P] [-q <pal_map> | -Q] [-s <nb_colors>]\n"
+	    "       [-t <tile_map> | -T] [-x <nb_tiles>] <file>\n"
 	    "Useful options:\n"
 	    "    -m, --mirror-tiles    optimize out mirrored tiles\n"
 	    "    -o, --output <path>   output the tile data to this path\n"
@@ -159,10 +175,8 @@ static void printUsage() {
 	);
 }
 
-/*
- * Parses a number at the beginning of a string, moving the pointer to skip the parsed characters
- * Returns the provided errVal on error
- */
+// Parses a number at the beginning of a string, moving the pointer to skip the parsed characters.
+// Returns the provided errVal on error.
 static uint16_t parseNumber(char *&string, char const *errPrefix, uint16_t errVal = UINT16_MAX) {
 	uint8_t base = 10;
 	if (*string == '\0') {
@@ -185,12 +199,10 @@ static uint16_t parseNumber(char *&string, char const *errPrefix, uint16_t errVa
 		}
 	}
 
-	/*
-	 * Turns a digit into its numeric value in the current base, if it has one.
-	 * Maximum is inclusive. The string_view is modified to "consume" all digits.
-	 * Returns 255 on parse failure (including wrong char for base), in which case
-	 * the string_view may be pointing on garbage.
-	 */
+	// Turns a digit into its numeric value in the current base, if it has one.
+	// Maximum is inclusive. The string_view is modified to "consume" all digits.
+	// Returns 255 on parse failure (including wrong char for base), in which case
+	// the string_view may be pointing on garbage.
 	auto charIndex = [&base](unsigned char c) -> uint8_t {
 		unsigned char index = c - '0'; // Use wrapping semantics
 		if (base == 2 && index >= 2) {
@@ -258,10 +270,7 @@ static void registerInput(char const *arg) {
 	}
 }
 
-/*
- * Turn an "at-file"'s contents into an argv that `getopt` can handle
- * @param argPool Argument characters will be appended to this vector, for storage purposes.
- */
+// Turn an at-file's contents into an argv that `getopt` can handle, appending them to `argPool`.
 static std::vector<size_t> readAtFile(std::string const &path, std::vector<char> &argPool) {
 	File file;
 	if (!file.open(path, std::ios_base::in)) {
@@ -333,13 +342,10 @@ static std::vector<size_t> readAtFile(std::string const &path, std::vector<char>
 	}
 }
 
-/*
- * Parses an arg vector, modifying `options` and `localOptions` as options are read.
- * The `localOptions` struct is for flags which must be processed after the option parsing finishes.
- *
- * Returns `nullptr` if the vector was fully parsed, or a pointer (which is part of the arg vector)
- * to an "at-file" path if one is encountered.
- */
+// Parses an arg vector, modifying `options` and `localOptions` as options are read.
+// The `localOptions` struct is for flags which must be processed after the option parsing finishes.
+// Returns `nullptr` if the vector was fully parsed, or a pointer (which is part of the arg vector)
+// to an "at-file" path if one is encountered.
 static char *parseArgv(int argc, char *argv[]) {
 	for (int ch; (ch = musl_getopt_long_only(argc, argv, optstring, longopts, nullptr)) != -1;) {
 		char *arg = musl_optarg; // Make a copy for scanning
@@ -350,8 +356,9 @@ static char *parseArgv(int argc, char *argv[]) {
 			break;
 		case 'a':
 			localOptions.autoAttrmap = false;
-			if (!options.attrmap.empty())
+			if (!options.attrmap.empty()) {
 				warning("Overriding attrmap file %s", options.attrmap.c_str());
+			}
 			options.attrmap = musl_optarg;
 			break;
 		case 'b':
@@ -417,6 +424,15 @@ static char *parseArgv(int argc, char *argv[]) {
 				options.bitDepth = 2;
 			}
 			break;
+		case 'h':
+			printUsage();
+			exit(0);
+		case 'i':
+			if (!options.inputTileset.empty()) {
+				warning("Overriding input tileset file %s", options.inputTileset.c_str());
+			}
+			options.inputTileset = musl_optarg;
+			break;
 		case 'L':
 			options.inputSlice.left = parseNumber(arg, "Input slice left coordinate");
 			if (options.inputSlice.left > INT16_MAX) {
@@ -458,8 +474,9 @@ static char *parseArgv(int argc, char *argv[]) {
 			}
 			break;
 		case 'm':
-			options.allowMirroring = true;
-			[[fallthrough]]; // Imply `-u`
+			options.allowMirroringX = true; // Imply `-X`
+			options.allowMirroringY = true; // Imply `-Y`
+			[[fallthrough]];                // Imply `-u`
 		case 'u':
 			options.allowDedup = true;
 			break;
@@ -511,8 +528,9 @@ static char *parseArgv(int argc, char *argv[]) {
 			localOptions.groupOutputs = true;
 			break;
 		case 'o':
-			if (!options.output.empty())
+			if (!options.output.empty()) {
 				warning("Overriding tile data file %s", options.output.c_str());
+			}
 			options.output = musl_optarg;
 			break;
 		case 'P':
@@ -520,8 +538,9 @@ static char *parseArgv(int argc, char *argv[]) {
 			break;
 		case 'p':
 			localOptions.autoPalettes = false;
-			if (!options.palettes.empty())
+			if (!options.palettes.empty()) {
 				warning("Overriding palettes file %s", options.palettes.c_str());
+			}
 			options.palettes = musl_optarg;
 			break;
 		case 'Q':
@@ -529,18 +548,17 @@ static char *parseArgv(int argc, char *argv[]) {
 			break;
 		case 'q':
 			localOptions.autoPalmap = false;
-			if (!options.palmap.empty())
+			if (!options.palmap.empty()) {
 				warning("Overriding palette map file %s", options.palmap.c_str());
+			}
 			options.palmap = musl_optarg;
 			break;
 		case 'r':
+			localOptions.reverse = true;
 			options.reversedWidth = parseNumber(arg, "Reversed image stride");
 			if (*arg != '\0') {
 				error("Reversed image stride (-r) must be a valid number, not \"%s\"", musl_optarg);
 			}
-			if (options.reversedWidth == 0) {
-				error("Reversed image stride (-r) may not be 0!");
-			}
 			break;
 		case 's':
 			options.nbColorsPerPal = parseNumber(arg, "Number of colors per palette", 4);
@@ -558,8 +576,9 @@ static char *parseArgv(int argc, char *argv[]) {
 			break;
 		case 't':
 			localOptions.autoTilemap = false;
-			if (!options.tilemap.empty())
+			if (!options.tilemap.empty()) {
 				warning("Overriding tilemap file %s", options.tilemap.c_str());
+			}
 			options.tilemap = musl_optarg;
 			break;
 		case 'V':
@@ -576,6 +595,14 @@ static char *parseArgv(int argc, char *argv[]) {
 				error("Tile trim (-x) argument must be a valid number, not \"%s\"", musl_optarg);
 			}
 			break;
+		case 'X':
+			options.allowMirroringX = true;
+			options.allowDedup = true; // Imply `-u`
+			break;
+		case 'Y':
+			options.allowMirroringY = true;
+			options.allowDedup = true; // Imply `-u`
+			break;
 		case 'Z':
 			options.columnMajor = true;
 			break;
@@ -588,7 +615,6 @@ static char *parseArgv(int argc, char *argv[]) {
 			}
 			break;
 		default:
-			fprintf(stderr, "FATAL: unknown option '%c'\n", ch);
 			printUsage();
 			exit(1);
 		}
@@ -601,7 +627,6 @@ int main(int argc, char *argv[]) {
 	struct AtFileStackEntry {
 		int parentInd;            // Saved offset into parent argv
 		std::vector<char *> argv; // This context's arg pointer vec
-		std::vector<char> argPool;
 
 		AtFileStackEntry(int parentInd_, std::vector<char *> argv_)
 		    : parentInd(parentInd_), argv(argv_) {}
@@ -610,18 +635,24 @@ int main(int argc, char *argv[]) {
 
 	int curArgc = argc;
 	char **curArgv = argv;
+	std::vector<std::vector<char>> argPools;
 	for (;;) {
 		char *atFileName = parseArgv(curArgc, curArgv);
 		if (atFileName) {
+			// We need to allocate a new arg pool for each at-file, so as not to invalidate pointers
+			// previous at-files may have generated to their own arg pools.
+			// But for the same reason, the arg pool must also outlive the at-file's stack entry!
+			auto &argPool = argPools.emplace_back();
+
 			// Copy `argv[0]` for error reporting, and because option parsing skips it
 			AtFileStackEntry &stackEntry =
 			    atFileStack.emplace_back(musl_optind, std::vector{atFileName});
 			// It would be nice to compute the char pointers on the fly, but reallocs don't allow
 			// that; so we must compute the offsets after the pool is fixed
-			auto offsets = readAtFile(&musl_optarg[1], stackEntry.argPool);
+			auto offsets = readAtFile(&musl_optarg[1], argPool);
 			stackEntry.argv.reserve(offsets.size() + 2); // Avoid a bunch of reallocs
 			for (size_t ofs : offsets) {
-				stackEntry.argv.push_back(&stackEntry.argPool.data()[ofs]);
+				stackEntry.argv.push_back(&argPool.data()[ofs]);
 			}
 			stackEntry.argv.push_back(nullptr); // Don't forget the arg vector terminator!
 
@@ -745,18 +776,26 @@ int main(int argc, char *argv[]) {
 		}
 
 		fputs("Options:\n", stderr);
-		if (options.columnMajor)
+		if (options.columnMajor) {
 			fputs("\tVisit image in column-major order\n", stderr);
-		if (options.allowMirroring)
-			fputs("\tAllow mirroring tiles\n", stderr);
-		if (options.allowDedup)
+		}
+		if (options.allowDedup) {
 			fputs("\tAllow deduplicating tiles\n", stderr);
-		if (options.useColorCurve)
+		}
+		if (options.allowMirroringX) {
+			fputs("\tAllow deduplicating horizontally mirrored tiles\n", stderr);
+		}
+		if (options.allowMirroringY) {
+			fputs("\tAllow deduplicating vertically mirrored tiles\n", stderr);
+		}
+		if (options.useColorCurve) {
 			fputs("\tUse color curve\n", stderr);
+		}
 		fprintf(stderr, "\tBit depth: %" PRIu8 "bpp\n", options.bitDepth);
-		if (options.trim != 0)
+		if (options.trim != 0) {
 			fprintf(stderr, "\tTrim the last %" PRIu64 " tiles\n", options.trim);
-		fprintf(stderr, "\tMaximum %" PRIu8 " palettes\n", options.nbPalettes);
+		}
+		fprintf(stderr, "\tMaximum %" PRIu16 " palettes\n", options.nbPalettes);
 		fprintf(stderr, "\tPalettes contain %" PRIu8 " colors\n", options.nbColorsPerPal);
 		fprintf(stderr, "\t%s palette spec\n", [] {
 			switch (options.palSpecType) {
@@ -780,7 +819,7 @@ int main(int argc, char *argv[]) {
 						fputs("#none, ", stderr);
 					}
 				}
-				fputc('\n', stderr);
+				putc('\n', stderr);
 			}
 			fputs("\t]\n", stderr);
 		}
@@ -818,18 +857,17 @@ int main(int argc, char *argv[]) {
 		fputs("Ready.\n", stderr);
 	}
 
-	// Do not do anything if option parsing went wrong
-	if (nbErrors) {
-		giveUp();
-	}
+	// Do not do anything if option parsing went wrong.
+	requireZeroErrors();
 
 	if (!options.input.empty()) {
-		if (options.reverse()) {
+		if (localOptions.reverse) {
 			reverse();
 		} else {
 			process();
 		}
-	} else if (!options.palettes.empty() && options.palSpecType == Options::EXPLICIT && !options.reverse()) {
+	} else if (!options.palettes.empty() && options.palSpecType == Options::EXPLICIT
+	           && !localOptions.reverse) {
 		processPalettes();
 	} else {
 		fputs("FATAL: No input image specified\n", stderr);
@@ -837,9 +875,7 @@ int main(int argc, char *argv[]) {
 		exit(1);
 	}
 
-	if (nbErrors) {
-		giveUp();
-	}
+	requireZeroErrors();
 	return 0;
 }
 
@@ -855,9 +891,7 @@ void Palette::addColor(uint16_t color) {
 	}
 }
 
-/*
- * Returns the ID of the color in the palette, or `size()` if the color is not in
- */
+// Returns the ID of the color in the palette, or `size()` if the color is not in
 uint8_t Palette::indexOf(uint16_t color) const {
 	return color == Rgba::transparent
 	           ? 0
