@@ -1,24 +1,23 @@
-/* SPDX-License-Identifier: MIT */
+// SPDX-License-Identifier: MIT
 
 #include "gfx/pal_packing.hpp"
 
 #include <algorithm>
 #include <deque>
 #include <inttypes.h>
+#include <numeric>
 #include <optional>
 #include <queue>
+#include <stdint.h>
 #include <type_traits>
 #include <unordered_set>
+#include <utility>
 
 #include "helpers.hpp"
 
 #include "gfx/main.hpp"
 #include "gfx/proto_palette.hpp"
 
-using std::swap;
-
-namespace packing {
-
 // The solvers here are picked from the paper at https://arxiv.org/abs/1605.00558:
 // "Algorithms for the Pagination Problem, a Bin Packing with Overlapping Items"
 // Their formulation of the problem consists in packing "tiles" into "pages"; here is a
@@ -28,15 +27,11 @@ namespace packing {
 //  Tile | Proto-palette
 //  Page | Palette
 
-/*
- * A reference to a proto-palette, and attached attributes for sorting purposes
- */
+// A reference to a proto-palette, and attached attributes for sorting purposes
 struct ProtoPalAttrs {
-	size_t const protoPalIndex;
-	/*
-	 * Pages from which we are banned (to prevent infinite loops)
-	 * This is dynamic because we wish not to hard-cap the amount of palettes
-	 */
+	size_t protoPalIndex;
+	// Pages from which we are banned (to prevent infinite loops)
+	// This is dynamic because we wish not to hard-cap the amount of palettes
 	std::vector<bool> bannedPages;
 
 	explicit ProtoPalAttrs(size_t index) : protoPalIndex(index) {}
@@ -51,10 +46,8 @@ struct ProtoPalAttrs {
 	}
 };
 
-/*
- * A collection of proto-palettes assigned to a palette
- * Does not contain the actual color indices because we need to be able to remove elements
- */
+// A collection of proto-palettes assigned to a palette
+// Does not contain the actual color indices because we need to be able to remove elements
 class AssignedProtos {
 	// We leave room for emptied slots to avoid copying the structs around on removal
 	std::vector<std::optional<ProtoPalAttrs>> _assigned;
@@ -114,8 +107,8 @@ private:
 		}
 
 		friend void swap(Iter &lhs, Iter &rhs) {
-			swap(lhs._array, rhs._array);
-			swap(lhs._iter, rhs._iter);
+			std::swap(lhs._array, rhs._array);
+			std::swap(lhs._iter, rhs._iter);
 		}
 	};
 public:
@@ -128,10 +121,8 @@ public:
 	}
 	const_iterator end() const { return const_iterator{&_assigned, _assigned.end()}; }
 
-	/*
-	 * Assigns a new ProtoPalAttrs in a free slot, assuming there is one
-	 * Args are passed to the `ProtoPalAttrs`'s constructor
-	 */
+	// Assigns a new ProtoPalAttrs in a free slot, assuming there is one
+	// Args are passed to the `ProtoPalAttrs`'s constructor
 	template<typename... Ts>
 	void assign(Ts &&...args) {
 		auto freeSlot =
@@ -193,9 +184,7 @@ private:
 		return colors;
 	}
 public:
-	/*
-	 * Returns the number of distinct colors
-	 */
+	// Returns the number of distinct colors
 	size_t volume() const { return uniqueColors().size(); }
 	bool canFit(ProtoPalette const &protoPal) const {
 		auto &colors = uniqueColors();
@@ -203,28 +192,47 @@ public:
 		return colors.size() <= options.maxOpaqueColors();
 	}
 
-	/*
-	 * Computes the "relative size" of a proto-palette on this palette
-	 */
-	double relSizeOf(ProtoPalette const &protoPal) const {
+	// The `relSizeOf` method below should compute the sum, for each color in `protoPal`, of
+	// the reciprocal of the "multiplicity" of the color across "our" proto-palettes.
+	// However, literally computing the reciprocals would involve floating-point division, which
+	// leads to imprecision and even platform-specific differences.
+	// We avoid this by multiplying the reciprocals by a factor such that division always produces
+	// an integer; the LCM of all values the denominator can take is the smallest suitable factor.
+	static constexpr uint32_t scaleFactor = [] {
+		// Fold over 1..=17 with the associative LCM function
+		// (17 is the largest the denominator in `relSizeOf` below can be)
+		uint32_t factor = 1;
+		for (uint32_t n = 2; n <= 17; ++n) {
+			factor = std::lcm(factor, n);
+		}
+		return factor;
+	}();
+
+	// Computes the "relative size" of a proto-palette on this palette;
+	// it's a measure of how much this proto-palette would "cost" to introduce.
+	uint32_t relSizeOf(ProtoPalette const &protoPal) const {
 		// NOTE: this function must not call `uniqueColors`, or one of its callers will break!
-		double relSize = 0.;
+
+		uint32_t relSize = 0;
 		for (uint16_t color : protoPal) {
-			auto n = std::count_if(RANGE(*this), [this, &color](ProtoPalAttrs const &attrs) {
-				ProtoPalette const &pal = (*_protoPals)[attrs.protoPalIndex];
-				return std::find(RANGE(pal), color) != pal.end();
-			});
-			// NOTE: The paper and the associated code disagree on this: the code has
-			// this `1 +`, whereas the paper does not; its lack causes a division by 0
-			// if the symbol is not found anywhere, so I'm assuming the paper is wrong.
-			relSize += 1. / (1 + n);
+			auto multiplicity = // How many of our proto-palettes does this color also belong to?
+			    std::count_if(RANGE(*this), [this, &color](ProtoPalAttrs const &attrs) {
+				    ProtoPalette const &pal = (*_protoPals)[attrs.protoPalIndex];
+				    return std::find(RANGE(pal), color) != pal.end();
+			    });
+			// We increase the denominator by 1 here; the reference code does this,
+			// but the paper does not. Not adding 1 makes a multiplicity of 0 cause a division by 0
+			// (that is, if the color is not found in any proto-palette), and adding 1 still seems
+			// to preserve the paper's reasoning.
+			//
+			// The scale factor should ensure integer divisions only.
+			assume(scaleFactor % (multiplicity + 1) == 0);
+			relSize += scaleFactor / (multiplicity + 1);
 		}
 		return relSize;
 	}
 
-	/*
-	 * Computes the "relative size" of a set of proto-palettes on this palette
-	 */
+	// Computes the "relative size" of a set of proto-palettes on this palette
 	template<typename Iter>
 	auto combinedVolume(Iter &&begin, Iter const &end, std::vector<ProtoPalette> const &protoPals)
 	    const {
@@ -232,9 +240,7 @@ public:
 		addUniqueColors(colors, std::forward<Iter>(begin), end, protoPals);
 		return colors.size();
 	}
-	/*
-	 * Computes the "relative size" of a set of colors on this palette
-	 */
+	// Computes the "relative size" of a set of colors on this palette
 	template<typename Iter>
 	auto combinedVolume(Iter &&begin, Iter &&end) const {
 		auto &colors = uniqueColors();
@@ -359,11 +365,19 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 	);
 
 	// Sort the proto-palettes by size, which improves the packing algorithm's efficiency
+	auto const indexOfLargestProtoPalFirst = [&protoPalettes](size_t left, size_t right) {
+		ProtoPalette const &lhs = protoPalettes[left];
+		ProtoPalette const &rhs = protoPalettes[right];
+		return lhs.size() > rhs.size(); // We want the proto-pals to be sorted *largest first*!
+	};
 	DefaultInitVec<size_t> sortedProtoPalIDs(protoPalettes.size());
 	sortedProtoPalIDs.clear();
 	for (size_t i = 0; i < protoPalettes.size(); ++i) {
-		sortedProtoPalIDs.insert(std::lower_bound(RANGE(sortedProtoPalIDs), i), i);
+		sortedProtoPalIDs.insert(
+		    std::lower_bound(RANGE(sortedProtoPalIDs), i, indexOfLargestProtoPalFirst), i
+		);
 	}
+
 	// Begin with all proto-palettes queued up for insertion
 	std::queue<ProtoPalAttrs> queue(std::deque<ProtoPalAttrs>(RANGE(sortedProtoPalIDs)));
 	// Begin with no pages
@@ -371,13 +385,15 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 
 	for (; !queue.empty(); queue.pop()) {
 		ProtoPalAttrs const &attrs = queue.front(); // Valid until the `queue.pop()`
-		options.verbosePrint(Options::VERB_DEBUG, "Handling proto-pal %zu\n", attrs.protoPalIndex);
+		options.verbosePrint(
+		    Options::VERB_TRACE, "Handling proto-palette %zu\n", attrs.protoPalIndex
+		);
 
 		ProtoPalette const &protoPal = protoPalettes[attrs.protoPalIndex];
 		size_t bestPalIndex = assignments.size();
 		// We're looking for a palette where the proto-palette's relative size is less than
 		// its actual size; so only overwrite the "not found" index on meeting that criterion
-		double bestRelSize = protoPal.size();
+		uint32_t bestRelSize = protoPal.size() * AssignedProtos::scaleFactor;
 
 		for (size_t i = 0; i < assignments.size(); ++i) {
 			// Skip the page if this one is banned from it
@@ -385,23 +401,37 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 				continue;
 			}
 
+			uint32_t relSize = assignments[i].relSizeOf(protoPal);
 			options.verbosePrint(
-			    Options::VERB_DEBUG,
-			    "%zu/%zu: Rel size: %f (size = %zu)\n",
-			    i + 1,
+			    Options::VERB_TRACE,
+			    "  Relative size to palette %zu (of %zu): %" PRIu32 " (size = %zu)\n",
+			    i,
 			    assignments.size(),
-			    assignments[i].relSizeOf(protoPal),
+			    relSize,
 			    protoPal.size()
 			);
-			if (assignments[i].relSizeOf(protoPal) < bestRelSize) {
+			if (relSize < bestRelSize) {
 				bestPalIndex = i;
+				bestRelSize = relSize;
 			}
 		}
 
 		if (bestPalIndex == assignments.size()) {
 			// Found nowhere to put it, create a new page containing just that one
+			options.verbosePrint(
+			    Options::VERB_TRACE,
+			    "Assigning proto-palette %zu to new palette %zu\n",
+			    attrs.protoPalIndex,
+			    bestPalIndex
+			);
 			assignments.emplace_back(protoPalettes, std::move(attrs));
 		} else {
+			options.verbosePrint(
+			    Options::VERB_TRACE,
+			    "Assigning proto-palette %zu to palette %zu\n",
+			    attrs.protoPalIndex,
+			    bestPalIndex
+			);
 			auto &bestPal = assignments[bestPalIndex];
 			// Add the color to that palette
 			bestPal.assign(std::move(attrs));
@@ -409,7 +439,7 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 			// If this overloads the palette, get it back to normal (if possible)
 			while (bestPal.volume() > options.maxOpaqueColors()) {
 				options.verbosePrint(
-				    Options::VERB_DEBUG,
+				    Options::VERB_TRACE,
 				    "Palette %zu is overloaded! (%zu > %" PRIu8 ")\n",
 				    bestPalIndex,
 				    bestPal.volume(),
@@ -417,28 +447,67 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 				);
 
 				// Look for a proto-pal minimizing "efficiency" (size / rel_size)
-				auto efficiency = [&bestPal](ProtoPalette const &pal) {
-					return pal.size() / bestPal.relSizeOf(pal);
-				};
 				auto [minEfficiencyIter, maxEfficiencyIter] = std::minmax_element(
 				    RANGE(bestPal),
-				    [&efficiency,
-				     &protoPalettes](ProtoPalAttrs const &lhs, ProtoPalAttrs const &rhs) {
-					    return efficiency(protoPalettes[lhs.protoPalIndex])
-					           < efficiency(protoPalettes[rhs.protoPalIndex]);
+				    [&bestPal, &protoPalettes](ProtoPalAttrs const &lhs, ProtoPalAttrs const &rhs) {
+					    ProtoPalette const &lhsProtoPal = protoPalettes[lhs.protoPalIndex];
+					    ProtoPalette const &rhsProtoPal = protoPalettes[rhs.protoPalIndex];
+					    size_t lhsSize = lhsProtoPal.size();
+					    size_t rhsSize = rhsProtoPal.size();
+					    uint32_t lhsRelSize = bestPal.relSizeOf(lhsProtoPal);
+					    uint32_t rhsRelSize = bestPal.relSizeOf(rhsProtoPal);
+
+					    options.verbosePrint(
+					        Options::VERB_TRACE,
+					        "  Proto-palettes %zu <=> %zu: Efficiency: %zu / %" PRIu32 " <=> %zu / "
+					        "%" PRIu32 "\n",
+					        lhs.protoPalIndex,
+					        rhs.protoPalIndex,
+					        lhsSize,
+					        lhsRelSize,
+					        rhsSize,
+					        rhsRelSize
+					    );
+					    // This comparison is algebraically equivalent to
+					    // `lhsSize / lhsRelSize < rhsSize / rhsRelSize`,
+					    // but without potential precision loss from floating-point division.
+					    return lhsSize * rhsRelSize < rhsSize * lhsRelSize;
 				    }
 				);
 
 				// All efficiencies are identical iff min equals max
 				// TODO: maybe not ideal to re-compute these two?
-				// TODO: yikes for float comparison! I *think* this threshold is OK?
-				if (efficiency(protoPalettes[maxEfficiencyIter->protoPalIndex])
-				        - efficiency(protoPalettes[minEfficiencyIter->protoPalIndex])
-				    < .001) {
+				ProtoPalette const &minProtoPal = protoPalettes[minEfficiencyIter->protoPalIndex];
+				ProtoPalette const &maxProtoPal = protoPalettes[maxEfficiencyIter->protoPalIndex];
+				size_t minSize = minProtoPal.size();
+				size_t maxSize = maxProtoPal.size();
+				uint32_t minRelSize = bestPal.relSizeOf(minProtoPal);
+				uint32_t maxRelSize = bestPal.relSizeOf(maxProtoPal);
+				options.verbosePrint(
+				    Options::VERB_TRACE,
+				    "  Proto-palettes %zu <= %zu: Efficiency: %zu / %" PRIu32 " <= %zu / %" PRIu32
+				    "\n",
+				    minEfficiencyIter->protoPalIndex,
+				    maxEfficiencyIter->protoPalIndex,
+				    minSize,
+				    minRelSize,
+				    maxSize,
+				    maxRelSize
+				);
+				// This comparison is algebraically equivalent to
+				// `maxSize / maxRelSize == minSize / minRelSize`,
+				// but without potential precision loss from floating-point division.
+				if (maxSize * minRelSize == minSize * maxRelSize) {
+					options.verbosePrint(Options::VERB_TRACE, "  All efficiencies are identical\n");
 					break;
 				}
 
 				// Remove the proto-pal with minimal efficiency
+				options.verbosePrint(
+				    Options::VERB_TRACE,
+				    "  Removing proto-palette %zu\n",
+				    minEfficiencyIter->protoPalIndex
+				);
 				queue.emplace(std::move(*minEfficiencyIter));
 				queue.back().banFrom(bestPalIndex); // Ban it from this palette
 				bestPal.remove(minEfficiencyIter);
@@ -447,17 +516,25 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 	}
 
 	// Deal with palettes still overloaded, by emptying them
+	auto const &largestProtoPalFirst =
+	    [&protoPalettes](ProtoPalAttrs const &lhs, ProtoPalAttrs const &rhs) {
+		    return protoPalettes[lhs.protoPalIndex].size()
+		           > protoPalettes[rhs.protoPalIndex].size();
+	    };
+	std::vector<ProtoPalAttrs> overloadQueue{};
 	for (AssignedProtos &pal : assignments) {
 		if (pal.volume() > options.maxOpaqueColors()) {
 			for (ProtoPalAttrs &attrs : pal) {
-				queue.emplace(std::move(attrs));
+				overloadQueue.emplace(
+				    std::lower_bound(RANGE(overloadQueue), attrs, largestProtoPalFirst),
+				    std::move(attrs)
+				);
 			}
 			pal.clear();
 		}
 	}
 	// Place back any proto-palettes now in the queue via first-fit
-	while (!queue.empty()) {
-		ProtoPalAttrs const &attrs = queue.front();
+	for (ProtoPalAttrs const &attrs : overloadQueue) {
 		ProtoPalette const &protoPal = protoPalettes[attrs.protoPalIndex];
 		auto iter = std::find_if(RANGE(assignments), [&protoPal](AssignedProtos const &pal) {
 			return pal.canFit(protoPal);
@@ -465,7 +542,7 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 		if (iter == assignments.end()) { // No such page, create a new one
 			options.verbosePrint(
 			    Options::VERB_DEBUG,
-			    "Adding new palette (%zu) for overflowing proto-pal %zu\n",
+			    "Adding new palette (%zu) for overflowing proto-palette %zu\n",
 			    assignments.size(),
 			    attrs.protoPalIndex
 			);
@@ -473,13 +550,12 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 		} else {
 			options.verbosePrint(
 			    Options::VERB_DEBUG,
-			    "Assigning overflowing proto-pal %zu to palette %zu\n",
+			    "Assigning overflowing proto-palette %zu to palette %zu\n",
 			    attrs.protoPalIndex,
 			    iter - assignments.begin()
 			);
 			iter->assign(std::move(attrs));
 		}
-		queue.pop();
 	}
 
 	if (options.verbosity >= Options::VERB_INTERM) {
@@ -520,5 +596,3 @@ std::tuple<DefaultInitVec<size_t>, size_t>
 	}
 	return {mappings, assignments.size()};
 }
-
-} // namespace packing
