@@ -1,4 +1,4 @@
-/* SPDX-License-Identifier: MIT */
+// SPDX-License-Identifier: MIT
 
 #include "gfx/process.hpp"
 
@@ -15,7 +15,7 @@
 #include <utility>
 #include <vector>
 
-#include "defaultinitalloc.hpp"
+#include "defaultinitvec.hpp"
 #include "file.hpp"
 #include "helpers.hpp"
 #include "itertools.hpp"
@@ -26,18 +26,16 @@
 #include "gfx/proto_palette.hpp"
 
 class ImagePalette {
-	// Use as many slots as there are CGB colors (plus transparency)
-	std::array<std::optional<Rgba>, 0x8001> _colors;
+	std::array<std::optional<Rgba>, NB_COLOR_SLOTS> _colors;
 
 public:
 	ImagePalette() = default;
 
-	/*
-	 * Registers a color in the palette.
-	 * If the newly inserted color "conflicts" with another one (different color, but same CGB
-	 * color), then the other color is returned. Otherwise, `nullptr` is returned.
-	 */
-	[[nodiscard]] Rgba const *registerColor(Rgba const &rgba) {
+	// Registers a color in the palette.
+	// If the newly inserted color "conflicts" with another one (different color, but same CGB
+	// color), then the other color is returned. Otherwise, `nullptr` is returned.
+	[[nodiscard]]
+	Rgba const *registerColor(Rgba const &rgba) {
 		decltype(_colors)::value_type &slot = _colors[rgba.cgbColor()];
 
 		if (rgba.cgbColor() == Rgba::transparent) {
@@ -80,7 +78,8 @@ class Png {
 	int nbTransparentEntries;
 	png_bytep transparencyPal = nullptr;
 
-	[[noreturn]] static void handleError(png_structp png, char const *msg) {
+	[[noreturn]]
+	static void handleError(png_structp png, char const *msg) {
 		Png *self = reinterpret_cast<Png *>(png_get_error_ptr(png));
 
 		fatal("Error reading input image (\"%s\"): %s", self->c_str(), msg);
@@ -104,7 +103,7 @@ class Png {
 			    "bytes after reading %zu)",
 			    self->c_str(),
 			    length - nbBytesRead,
-			    (size_t)self->file->pubseekoff(0, std::ios_base::cur)
+			    static_cast<size_t>(self->file->pubseekoff(0, std::ios_base::cur))
 			);
 		}
 	}
@@ -166,15 +165,13 @@ public:
 		return true;
 	}
 
-	/*
-	 * Reads a PNG and notes all of its colors
-	 *
-	 * This code is more complicated than strictly necessary, but that's because of the API
-	 * being used: the "high-level" interface doesn't provide all the transformations we need,
-	 * so we use the "lower-level" one instead.
-	 * We also use that occasion to only read the PNG one line at a time, since we store all of
-	 * the pixel data in `pixels`, which saves on memory allocations.
-	 */
+	// Reads a PNG and notes all of its colors
+	//
+	// This code is more complicated than strictly necessary, but that's because of the API
+	// being used: the "high-level" interface doesn't provide all the transformations we need,
+	// so we use the "lower-level" one instead.
+	// We also use that occasion to only read the PNG one line at a time, since we store all of
+	// the pixel data in `pixels`, which saves on memory allocations.
 	explicit Png(std::string const &filePath) : path(filePath), colors() {
 		if (file.open(path, std::ios_base::in | std::ios_base::binary) == nullptr) {
 			fatal("Failed to open input image (\"%s\"): %s", file.c_str(path), strerror(errno));
@@ -193,7 +190,7 @@ public:
 		options.verbosePrint(Options::VERB_INTERM, "PNG header signature is OK\n");
 
 		png = png_create_read_struct(
-		    PNG_LIBPNG_VER_STRING, (png_voidp)this, handleError, handleWarning
+		    PNG_LIBPNG_VER_STRING, static_cast<png_voidp>(this), handleError, handleWarning
 		);
 		if (!png) {
 			fatal("Failed to allocate PNG structure: %s", strerror(errno));
@@ -446,7 +443,7 @@ public:
 		};
 
 	private:
-		struct iterator {
+		struct Iterator {
 			TilesVisitor const &parent;
 			uint32_t const limit;
 			uint32_t x, y;
@@ -458,7 +455,7 @@ public:
 				return {parent._png, x + options.inputSlice.left, y + options.inputSlice.top};
 			}
 
-			iterator &operator++() {
+			Iterator &operator++() {
 				auto [major, minor] = parent._columnMajor ? std::tie(y, x) : std::tie(x, y);
 				major += 8;
 				if (major == limit) {
@@ -468,19 +465,14 @@ public:
 				return *this;
 			}
 
-			friend bool operator==(iterator const &lhs, iterator const &rhs) {
-				return lhs.coords() == rhs.coords(); // Compare the returned coord pairs
-			}
-
-			friend bool operator!=(iterator const &lhs, iterator const &rhs) {
-				return lhs.coords() != rhs.coords(); // Compare the returned coord pairs
-			}
+			bool operator==(Iterator const &rhs) const { return coords() == rhs.coords(); }
+			bool operator!=(Iterator const &rhs) const { return coords() != rhs.coords(); }
 		};
 
 	public:
-		iterator begin() const { return {*this, _limit, 0, 0}; }
-		iterator end() const {
-			iterator it{*this, _limit, _width - 8, _height - 8}; // Last valid one...
+		Iterator begin() const { return {*this, _limit, 0, 0}; }
+		Iterator end() const {
+			Iterator it{*this, _limit, _width - 8, _height - 8}; // Last valid one...
 			return ++it;                                         // ...now one-past-last!
 		}
 	};
@@ -496,9 +488,7 @@ public:
 };
 
 class RawTiles {
-	/*
-	 * A tile which only contains indices into the image's global palette
-	 */
+	// A tile which only contains indices into the image's global palette
 	class RawTile {
 		std::array<std::array<size_t, 8>, 8> _pixelIndices{};
 
@@ -511,9 +501,7 @@ private:
 	std::vector<RawTile> _tiles;
 
 public:
-	/*
-	 * Creates a new raw tile, and returns a reference to it so it can be filled in
-	 */
+	// Creates a new raw tile, and returns a reference to it so it can be filled in
 	RawTile &newTile() {
 		_tiles.emplace_back();
 		return _tiles.back();
@@ -521,11 +509,9 @@ public:
 };
 
 struct AttrmapEntry {
-	/*
-	 * This field can either be a proto-palette ID, or `transparent` to indicate that the
-	 * corresponding tile is fully transparent. If you are looking to get the palette ID for this
-	 * attrmap entry while correctly handling the above, use `getPalID`.
-	 */
+	// This field can either be a proto-palette ID, or `transparent` to indicate that the
+	// corresponding tile is fully transparent. If you are looking to get the palette ID for this
+	// attrmap entry while correctly handling the above, use `getPalID`.
 	size_t protoPaletteID; // Only this field is used when outputting "unoptimized" data
 	uint8_t tileID;        // This is the ID as it will be output to the tilemap
 	bool bank;
@@ -567,7 +553,7 @@ static std::tuple<DefaultInitVec<size_t>, std::vector<Palette>>
     generatePalettes(std::vector<ProtoPalette> const &protoPalettes, Png const &png) {
 	// Run a "pagination" problem solver
 	// TODO: allow picking one of several solvers?
-	auto [mappings, nbPalettes] = packing::overloadAndRemove(protoPalettes);
+	auto [mappings, nbPalettes] = overloadAndRemove(protoPalettes);
 	assume(mappings.size() == protoPalettes.size());
 
 	if (options.verbosity >= Options::VERB_INTERM) {
@@ -601,11 +587,11 @@ static std::tuple<DefaultInitVec<size_t>, std::vector<Palette>>
 	// "Sort" colors in the generated palettes, see the man page for the flowchart
 	auto [embPalSize, embPalRGB, embPalAlphaSize, embPalAlpha] = png.getEmbeddedPal();
 	if (embPalRGB != nullptr) {
-		sorting::indexed(palettes, embPalSize, embPalRGB, embPalAlphaSize, embPalAlpha);
+		sortIndexed(palettes, embPalSize, embPalRGB, embPalAlphaSize, embPalAlpha);
 	} else if (png.isSuitableForGrayscale()) {
-		sorting::grayscale(palettes, png.getColors().raw());
+		sortGrayscale(palettes, png.getColors().raw());
 	} else {
-		sorting::rgb(palettes);
+		sortRgb(palettes);
 	}
 	return {mappings, palettes};
 }
@@ -627,10 +613,9 @@ static std::tuple<DefaultInitVec<size_t>, std::vector<Palette>>
 		static char buf[sizeof(", $XXXX, $XXXX, $XXXX, $XXXX")];
 		char *ptr = buf;
 		for (uint16_t cgbColor : list) {
-			sprintf(ptr, ", $%04x", cgbColor);
-			ptr += QUOTEDSTRLEN(", $XXXX");
+			ptr += snprintf(ptr, sizeof(", $XXXX"), ", $%04x", cgbColor);
 		}
-		return &buf[QUOTEDSTRLEN(", ")];
+		return &buf[literal_strlen(", ")];
 	};
 
 	// Iterate through proto-palettes, and try mapping them to the specified palettes
@@ -683,7 +668,7 @@ static void outputPalettes(std::vector<Palette> const &palettes) {
 		// If the palette generation is wrong, other (dependee) operations are likely to be
 		// nonsensical, so fatal-error outright
 		fatal(
-		    "Generated %zu palettes, over the maximum of %" PRIu8,
+		    "Generated %zu palettes, over the maximum of %" PRIu16,
 		    palettes.size(),
 		    options.nbPalettes
 		);
@@ -692,7 +677,7 @@ static void outputPalettes(std::vector<Palette> const &palettes) {
 	if (!options.palettes.empty()) {
 		File output;
 		if (!output.open(options.palettes, std::ios_base::out | std::ios_base::binary)) {
-			fatal("Failed to open \"%s\": %s", output.c_str(options.palettes), strerror(errno));
+			fatal("Failed to create \"%s\": %s", output.c_str(options.palettes), strerror(errno));
 		}
 
 		for (Palette const &palette : palettes) {
@@ -706,7 +691,22 @@ static void outputPalettes(std::vector<Palette> const &palettes) {
 	}
 }
 
+static void hashBitplanes(uint16_t bitplanes, uint16_t &hash) {
+	hash ^= bitplanes;
+	if (options.allowMirroringX) {
+		// Count the line itself as mirrored, which ensures the same hash as the tile's horizontal
+		// flip; vertical mirroring is already taken care of because the symmetric line will be
+		// XOR'd the same way. (This can trivially create some collisions, but real-world tile data
+		// generally doesn't trigger them.)
+		hash ^= flipTable[bitplanes >> 8] << 8 | flipTable[bitplanes & 0xFF];
+	}
+}
+
 class TileData {
+	// Importantly, `TileData` is **always** 2bpp.
+	// If the active bit depth is 1bpp, all tiles are processed as 2bpp nonetheless, but emitted as
+	// 1bpp. This massively simplifies internal processing, since bit depth is always identical
+	// outside of I/O / serialization boundaries.
 	std::array<uint8_t, 16> _data;
 	// The hash is a bit lax: it's the XOR of all lines, and every other nibble is identical
 	// if horizontal mirroring is in effect. It should still be a reasonable tie-breaker in
@@ -736,23 +736,21 @@ public:
 		return row;
 	}
 
+	TileData(std::array<uint8_t, 16> &&raw) : _data(raw), _hash(0) {
+		for (uint8_t y = 0; y < 8; ++y) {
+			uint16_t bitplanes = _data[y * 2] | _data[y * 2 + 1] << 8;
+			hashBitplanes(bitplanes, _hash);
+		}
+	}
+
 	TileData(Png::TilesVisitor::Tile const &tile, Palette const &palette) : _hash(0) {
 		size_t writeIndex = 0;
 		for (uint32_t y = 0; y < 8; ++y) {
 			uint16_t bitplanes = rowBitplanes(tile, palette, y);
-			_data[writeIndex++] = bitplanes & 0xFF;
-			if (options.bitDepth == 2) {
-				_data[writeIndex++] = bitplanes >> 8;
-			}
+			hashBitplanes(bitplanes, _hash);
 
-			// Update the hash
-			_hash ^= bitplanes;
-			if (options.allowMirroring) {
-				// Count the line itself as mirrorred; vertical mirroring is
-				// already taken care of because the symmetric line will be XOR'd
-				// the same way. (...which is a problem, but probably benign.)
-				_hash ^= flipTable[bitplanes >> 8] << 8 | flipTable[bitplanes & 0xFF];
-			}
+			_data[writeIndex++] = bitplanes & 0xFF;
+			_data[writeIndex++] = bitplanes >> 8;
 		}
 	}
 
@@ -773,15 +771,17 @@ public:
 			return MatchType::EXACT;
 		}
 
-		if (!options.allowMirroring) {
-			return MatchType::NOPE;
+		// Check if we have horizontal mirroring, which scans the array forward again
+		if (options.allowMirroringX
+		    && std::equal(RANGE(_data), other._data.begin(), [](uint8_t lhs, uint8_t rhs) {
+			       return lhs == flipTable[rhs];
+		       })) {
+			return MatchType::HFLIP;
 		}
 
-		// Check if we have horizontal mirroring, which scans the array forward again
-		if (std::equal(RANGE(_data), other._data.begin(), [](uint8_t lhs, uint8_t rhs) {
-			    return lhs == flipTable[rhs];
-		    })) {
-			return MatchType::HFLIP;
+		// The remaining possibilities for matching all require vertical mirroring
+		if (!options.allowMirroringY) {
+			return MatchType::NOPE;
 		}
 
 		// Check if we have vertical or vertical+horizontal mirroring, for which we have to read
@@ -803,12 +803,18 @@ public:
 		}
 
 		// If we have both (i.e. we have symmetry), default to vflip only
-		assume(hasVFlip || hasVHFlip);
-		return hasVFlip ? MatchType::VFLIP : MatchType::VHFLIP;
-	}
-	friend bool operator==(TileData const &lhs, TileData const &rhs) {
-		return lhs.tryMatching(rhs) != MatchType::NOPE;
+		if (hasVFlip) {
+			return MatchType::VFLIP;
+		}
+
+		// If we allow both and have both, then use both
+		if (options.allowMirroringX && hasVHFlip) {
+			return MatchType::VHFLIP;
+		}
+
+		return MatchType::NOPE;
 	}
+	bool operator==(TileData const &rhs) const { return tryMatching(rhs) != MatchType::NOPE; }
 };
 
 template<>
@@ -816,9 +822,7 @@ struct std::hash<TileData> {
 	std::size_t operator()(TileData const &tile) const { return tile.hash(); }
 };
 
-namespace unoptimized {
-
-static void outputTileData(
+static void outputUnoptimizedTileData(
     Png const &png,
     DefaultInitVec<AttrmapEntry> const &attrmap,
     std::vector<Palette> const &palettes,
@@ -826,7 +830,7 @@ static void outputTileData(
 ) {
 	File output;
 	if (!output.open(options.output, std::ios_base::out | std::ios_base::binary)) {
-		fatal("Failed to open \"%s\": %s", output.c_str(options.output), strerror(errno));
+		fatal("Failed to create \"%s\": %s", output.c_str(options.output), strerror(errno));
 	}
 
 	uint16_t widthTiles = options.inputSlice.width ? options.inputSlice.width : png.getWidth() / 8;
@@ -857,7 +861,7 @@ static void outputTileData(
 	assume(remainingTiles == 0);
 }
 
-static void outputMaps(
+static void outputUnoptimizedMaps(
     DefaultInitVec<AttrmapEntry> const &attrmap, DefaultInitVec<size_t> const &mappings
 ) {
 	std::optional<File> tilemapOutput, attrmapOutput, palmapOutput;
@@ -865,7 +869,7 @@ static void outputMaps(
 		if (!path.empty()) {
 			file.emplace();
 			if (!file->open(path, std::ios_base::out | std::ios_base::binary)) {
-				fatal("Failed to open \"%s\": %s", file->c_str(options.tilemap), strerror(errno));
+				fatal("Failed to create \"%s\": %s", file->c_str(options.tilemap), strerror(errno));
 			}
 		}
 	};
@@ -896,10 +900,6 @@ static void outputMaps(
 	}
 }
 
-} // namespace unoptimized
-
-namespace optimized {
-
 struct UniqueTiles {
 	std::unordered_set<TileData> tileset;
 	std::vector<TileData const *> tiles;
@@ -910,15 +910,11 @@ struct UniqueTiles {
 	UniqueTiles(UniqueTiles const &) = delete;
 	UniqueTiles(UniqueTiles &&) = default;
 
-	/*
-	 * Adds a tile to the collection, and returns its ID
-	 */
-	std::tuple<uint16_t, TileData::MatchType>
-	    addTile(Png::TilesVisitor::Tile const &tile, Palette const &palette) {
-		TileData newTile(tile, palette);
+	// Adds a tile to the collection, and returns its ID
+	std::tuple<uint16_t, TileData::MatchType> addTile(TileData newTile) {
 		auto [tileData, inserted] = tileset.insert(newTile);
 
-		TileData::MatchType matchType = TileData::EXACT;
+		TileData::MatchType matchType = TileData::NOPE;
 		if (inserted) {
 			// Give the new tile the next available unique ID
 			tileData->tileID = static_cast<uint16_t>(tiles.size());
@@ -936,12 +932,10 @@ struct UniqueTiles {
 	auto end() const { return tiles.end(); }
 };
 
-/*
- * Generate tile data while deduplicating unique tiles (via mirroring if enabled)
- * Additionally, while we have the info handy, convert from the 16-bit "global" tile IDs to
- * 8-bit tile IDs + the bank bit; this will save the work when we output the data later (potentially
- * twice)
- */
+// Generate tile data while deduplicating unique tiles (via mirroring if enabled)
+// Additionally, while we have the info handy, convert from the 16-bit "global" tile IDs to
+// 8-bit tile IDs + the bank bit; this will save the work when we output the data later (potentially
+// twice)
 static UniqueTiles dedupTiles(
     Png const &png,
     DefaultInitVec<AttrmapEntry> &attrmap,
@@ -953,8 +947,58 @@ static UniqueTiles dedupTiles(
 	// by caching the full tile data anyway, so we might as well.)
 	UniqueTiles tiles;
 
+	if (!options.inputTileset.empty()) {
+		File inputTileset;
+		if (!inputTileset.open(options.inputTileset, std::ios::in | std::ios::binary)) {
+			fatal("Failed to open \"%s\": %s", options.inputTileset.c_str(), strerror(errno));
+		}
+
+		std::array<uint8_t, 16> tile;
+		size_t const tileSize = options.bitDepth * 8;
+		for (;;) {
+			// It's okay to cast between character types.
+			size_t len = inputTileset->sgetn(reinterpret_cast<char *>(tile.data()), tileSize);
+			if (len == 0) { // EOF!
+				break;
+			} else if (len != tileSize) {
+				fatal(
+				    "\"%s\" does not contain a multiple of %zu bytes; is it actually tile data?",
+				    options.inputTileset.c_str(),
+				    tileSize
+				);
+			} else if (len == 8) {
+				// Expand the tile data to 2bpp.
+				for (size_t i = 8; i--;) {
+					tile[i * 2 + 1] = 0;
+					tile[i * 2] = tile[i];
+				}
+			}
+
+			auto [tileID, matchType] = tiles.addTile(std::move(tile));
+
+			if (matchType != TileData::NOPE) {
+				error(
+				    "The input tileset's tile #%hu was deduplicated; please check that your "
+				    "deduplication flags (`-u`, `-m`) are consistent with what was used to "
+				    "generate the input tileset",
+				    tileID
+				);
+			}
+		}
+	}
+
+	bool inputWithoutOutput = !options.inputTileset.empty() && options.output.empty();
 	for (auto [tile, attr] : zip(png.visitAsTiles(), attrmap)) {
-		auto [tileID, matchType] = tiles.addTile(tile, palettes[mappings[attr.protoPaletteID]]);
+		auto [tileID, matchType] = tiles.addTile({tile, palettes[mappings[attr.protoPaletteID]]});
+
+		if (inputWithoutOutput && matchType == TileData::NOPE) {
+			error(
+			    "Tile at (%" PRIu32 ", %" PRIu32
+			    ") is not within the input tileset, and `-o` was not given!",
+			    tile.x,
+			    tile.y
+			);
+		}
 
 		attr.xFlip = matchType == TileData::HFLIP || matchType == TileData::VHFLIP;
 		attr.yFlip = matchType == TileData::VFLIP || matchType == TileData::VHFLIP;
@@ -978,7 +1022,14 @@ static void outputTileData(UniqueTiles const &tiles) {
 		TileData const *tile = *iter;
 		assume(tile->tileID == tileID);
 		++tileID;
-		output->sputn(reinterpret_cast<char const *>(tile->data().data()), options.bitDepth * 8);
+		if (options.bitDepth == 2) {
+			output->sputn(reinterpret_cast<char const *>(tile->data().data()), 16);
+		} else {
+			assume(options.bitDepth == 1);
+			for (size_t y = 0; y < 8; ++y) {
+				output->sputc(tile->data()[y * 2]);
+			}
+		}
 	}
 }
 
@@ -1022,8 +1073,6 @@ static void outputPalmap(
 	}
 }
 
-} // namespace optimized
-
 void processPalettes() {
 	options.verbosePrint(Options::VERB_CFG, "Using libpng %s\n", png_get_libpng_ver(nullptr));
 
@@ -1063,47 +1112,57 @@ void process() {
 	DefaultInitVec<AttrmapEntry> attrmap{};
 
 	for (auto tile : png.visitAsTiles()) {
-		ProtoPalette tileColors;
 		AttrmapEntry &attrs = attrmap.emplace_back();
-		uint8_t nbColorsInTile = 0;
 
+		// Count the unique non-transparent colors for packing
+		std::unordered_set<uint16_t> tileColors;
 		for (uint32_t y = 0; y < 8; ++y) {
 			for (uint32_t x = 0; x < 8; ++x) {
-				Rgba color = tile.pixel(x, y);
-				if (!color.isTransparent()) { // Do not count transparency in for packing
-					// Add the color to the proto-pal (if not full), and count it if it was unique.
-					if (tileColors.add(color.cgbColor())) {
-						++nbColorsInTile;
-					}
+				if (Rgba color = tile.pixel(x, y); !color.isTransparent()) {
+					tileColors.insert(color.cgbColor());
 				}
 			}
 		}
 
+		if (tileColors.size() > options.maxOpaqueColors()) {
+			fatal(
+			    "Tile at (%" PRIu32 ", %" PRIu32 ") has %zu colors, more than %" PRIu8 "!",
+			    tile.x,
+			    tile.y,
+			    tileColors.size(),
+			    options.maxOpaqueColors()
+			);
+		}
+
 		if (tileColors.empty()) {
 			// "Empty" proto-palettes screw with the packing process, so discard those
 			attrs.protoPaletteID = AttrmapEntry::transparent;
 			continue;
 		}
 
+		ProtoPalette protoPalette;
+		for (uint16_t cgbColor : tileColors) {
+			protoPalette.add(cgbColor);
+		}
+
 		// Insert the proto-palette, making sure to avoid overlaps
 		for (size_t n = 0; n < protoPalettes.size(); ++n) {
-			switch (tileColors.compare(protoPalettes[n])) {
+			switch (protoPalette.compare(protoPalettes[n])) {
 			case ProtoPalette::WE_BIGGER:
-				protoPalettes[n] = tileColors; // Override them
+				protoPalettes[n] = protoPalette; // Override them
 				// Remove any other proto-palettes that we encompass
 				// (Example [(0, 1), (0, 2)], inserting (0, 1, 2))
-				/*
-				 * The following code does its job, except that references to the removed
-				 * proto-palettes are not updated, causing issues.
-				 * TODO: overlap might not be detrimental to the packing algorithm.
-				 * Investigation is necessary, especially if pathological cases are found.
-				 *
-				 * for (size_t i = protoPalettes.size(); --i != n;) {
-				 *     if (tileColors.compare(protoPalettes[i]) == ProtoPalette::WE_BIGGER) {
-				 *         protoPalettes.erase(protoPalettes.begin() + i);
-				 *     }
-				 * }
-				 */
+				//
+				// The following code does its job, except that references to the removed
+				// proto-palettes are not updated, causing issues.
+				// TODO: overlap might not be detrimental to the packing algorithm.
+				// Investigation is necessary, especially if pathological cases are found.
+				//
+				// for (size_t i = protoPalettes.size(); --i != n;) {
+				//     if (protoPalette.compare(protoPalettes[i]) == ProtoPalette::WE_BIGGER) {
+				//         protoPalettes.erase(protoPalettes.begin() + i);
+				//     }
+				// }
 				[[fallthrough]];
 
 			case ProtoPalette::THEY_BIGGER:
@@ -1116,17 +1175,6 @@ void process() {
 			}
 		}
 
-		if (nbColorsInTile > options.maxOpaqueColors()) {
-			fatal(
-			    "Tile at (%" PRIu32 ", %" PRIu32 ") has %" PRIu8 " opaque colors, more than %" PRIu8
-			    "!",
-			    tile.x,
-			    tile.y,
-			    nbColorsInTile,
-			    options.maxOpaqueColors()
-			);
-		}
-
 		attrs.protoPaletteID = protoPalettes.size();
 		if (protoPalettes.size() == AttrmapEntry::transparent) { // Check for overflow
 			fatal(
@@ -1134,7 +1182,7 @@ void process() {
 			    AttrmapEntry::transparent
 			);
 		}
-		protoPalettes.push_back(tileColors);
+		protoPalettes.push_back(protoPalette);
 continue_visiting_tiles:;
 	}
 
@@ -1176,9 +1224,15 @@ continue_visiting_tiles:;
 			);
 		}
 
+		// I currently cannot figure out useful semantics for this combination of flags.
+		if (!options.inputTileset.empty()) {
+			fatal("Input tilesets are not supported without `-u`\nPlease consider explaining your "
+			      "use case to RGBDS' developers!");
+		}
+
 		if (!options.output.empty()) {
 			options.verbosePrint(Options::VERB_LOG_ACT, "Generating unoptimized tile data...\n");
-			unoptimized::outputTileData(png, attrmap, palettes, mappings);
+			outputUnoptimizedTileData(png, attrmap, palettes, mappings);
 		}
 
 		if (!options.tilemap.empty() || !options.attrmap.empty() || !options.palmap.empty()) {
@@ -1186,12 +1240,12 @@ continue_visiting_tiles:;
 			    Options::VERB_LOG_ACT,
 			    "Generating unoptimized tilemap and/or attrmap and/or palmap...\n"
 			);
-			unoptimized::outputMaps(attrmap, mappings);
+			outputUnoptimizedMaps(attrmap, mappings);
 		}
 	} else {
 		// All of these require the deduplication process to be performed to be output
 		options.verbosePrint(Options::VERB_LOG_ACT, "Deduplicating tiles...\n");
-		optimized::UniqueTiles tiles = optimized::dedupTiles(png, attrmap, palettes, mappings);
+		UniqueTiles tiles = dedupTiles(png, attrmap, palettes, mappings);
 
 		if (tiles.size() > options.maxNbTiles[0] + options.maxNbTiles[1]) {
 			fatal(
@@ -1204,22 +1258,22 @@ continue_visiting_tiles:;
 
 		if (!options.output.empty()) {
 			options.verbosePrint(Options::VERB_LOG_ACT, "Generating optimized tile data...\n");
-			optimized::outputTileData(tiles);
+			outputTileData(tiles);
 		}
 
 		if (!options.tilemap.empty()) {
 			options.verbosePrint(Options::VERB_LOG_ACT, "Generating optimized tilemap...\n");
-			optimized::outputTilemap(attrmap);
+			outputTilemap(attrmap);
 		}
 
 		if (!options.attrmap.empty()) {
 			options.verbosePrint(Options::VERB_LOG_ACT, "Generating optimized attrmap...\n");
-			optimized::outputAttrmap(attrmap, mappings);
+			outputAttrmap(attrmap, mappings);
 		}
 
 		if (!options.palmap.empty()) {
 			options.verbosePrint(Options::VERB_LOG_ACT, "Generating optimized palmap...\n");
-			optimized::outputPalmap(attrmap, mappings);
+			outputPalmap(attrmap, mappings);
 		}
 	}
 }
