@@ -1,4 +1,4 @@
-/* SPDX-License-Identifier: MIT */
+// SPDX-License-Identifier: MIT
 
 #include "link/sdas_obj.hpp"
 
@@ -8,7 +8,6 @@
 #include <stdint.h>
 #include <string.h>
 #include <tuple>
-#include <variant>
 
 #include "helpers.hpp" // assume
 #include "linkdefs.hpp"
@@ -26,8 +25,9 @@ enum NumberType {
 };
 
 static void consumeLF(FileStackNode const &where, uint32_t lineNo, FILE *file) {
-	if (getc(file) != '\n')
+	if (getc(file) != '\n') {
 		fatal(&where, lineNo, "Bad line ending (CR without LF)");
+	}
 }
 
 static char const *delim = " \f\n\r\t\v"; // Whitespace according to the C and POSIX locales
@@ -37,6 +37,7 @@ static int
 retry:
 	++lineNo;
 	int firstChar = getc(file);
+	lineBuf.clear();
 
 	switch (firstChar) {
 	case EOF:
@@ -49,8 +50,9 @@ retry:
 		} while (firstChar != EOF && firstChar != '\r' && firstChar != '\n');
 		[[fallthrough]];
 	case '\r':
-		if (firstChar == '\r' && getc(file) != '\n')
+		if (firstChar == '\r' && getc(file) != '\n') {
 			consumeLF(where, lineNo, file);
+		}
 		[[fallthrough]];
 	case '\n':
 		goto retry;
@@ -90,14 +92,16 @@ static uint32_t readNumber(char const *str, char const *&endptr, NumberType base
 
 static uint32_t
     parseNumber(FileStackNode const &where, uint32_t lineNo, char const *str, NumberType base) {
-	if (str[0] == '\0')
+	if (str[0] == '\0') {
 		fatal(&where, lineNo, "Expected number, got empty string");
+	}
 
 	char const *endptr;
 	uint32_t res = readNumber(str, endptr, base);
 
-	if (*endptr != '\0')
+	if (*endptr != '\0') {
 		fatal(&where, lineNo, "Expected number, got \"%s\"", str);
+	}
 	return res;
 }
 
@@ -105,8 +109,9 @@ static uint8_t
     parseByte(FileStackNode const &where, uint32_t lineNo, char const *str, NumberType base) {
 	uint32_t num = parseNumber(where, lineNo, str, base);
 
-	if (num > UINT8_MAX)
+	if (num > UINT8_MAX) {
 		fatal(&where, lineNo, "\"%s\" is not a byte", str);
+	}
 	return num;
 }
 
@@ -136,25 +141,28 @@ enum RelocFlags {
 };
 
 void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol> &fileSymbols) {
-	std::vector<char> line(256);
+	std::vector<char> line;
+	line.reserve(256);
 	char const *token;
 
 #define getToken(ptr, ...) \
 	do { \
 		token = strtok((ptr), delim); \
-		if (!token) \
+		if (!token) { \
 			fatal(&where, lineNo, __VA_ARGS__); \
+		} \
 	} while (0)
 #define expectEol(...) \
 	do { \
 		token = strtok(nullptr, delim); \
-		if (token) \
+		if (token) { \
 			fatal(&where, lineNo, __VA_ARGS__); \
+		} \
 	} while (0)
 #define expectToken(expected, lineType) \
 	do { \
 		getToken(nullptr, "'%c' line is too short", (lineType)); \
-		if (strcasecmp(token, (expected)) != 0) \
+		if (strcasecmp(token, (expected)) != 0) { \
 			fatal( \
 			    &where, \
 			    lineNo, \
@@ -163,6 +171,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 			    (expected), \
 			    token \
 			); \
+		} \
 	} while (0)
 
 	uint32_t lineNo = 0;
@@ -200,18 +209,22 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 		fatal(&where, lineNo, "Unknown endianness type '%c'", line[0]);
 	}
 
-#define ADDR_SIZE 3
-	if (line[1] != '0' + ADDR_SIZE)
-		fatal(&where, lineNo, "Unknown or unsupported address size '%c'", line[1]);
+	static constexpr uint8_t ADDR_SIZE = 3;
 
-	if (line[2] != '\0')
+	if (line[1] != '0' + ADDR_SIZE) {
+		fatal(&where, lineNo, "Unknown or unsupported address size '%c'", line[1]);
+	}
+
+	if (line[2] != '\0') {
 		warning(&where, lineNo, "Ignoring unknown characters (\"%s\") in first line", &line[2]);
+	}
 
 	// Header line
 
 	lineType = nextLine(line, lineNo, where, file);
-	if (lineType != 'H')
+	if (lineType != 'H') {
 		fatal(&where, lineNo, "Expected header line, got '%c' line", lineType);
+	}
 	// Expected format: "A areas S global symbols"
 
 	getToken(line.data(), "Empty 'H' line");
@@ -221,6 +234,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 
 	getToken(nullptr, "'H' line is too short");
 	uint32_t expectedNbSymbols = parseNumber(where, lineNo, token, numberType);
+	fileSymbols.reserve(expectedNbSymbols);
 
 	expectToken("global", 'H');
 
@@ -239,8 +253,9 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 
 	for (;;) {
 		lineType = nextLine(line, lineNo, where, file);
-		if (lineType == EOF)
+		if (lineType == EOF) {
 			break;
+		}
 		switch (lineType) {
 		case 'M': // Module name
 		case 'O': // Assembler flags
@@ -248,18 +263,23 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 			break;
 
 		case 'A': {
-			if (fileSections.size() == expectedNbAreas)
+			if (fileSections.size() == expectedNbAreas) {
 				warning(
 				    &where, lineNo, "Got more 'A' lines than the expected %" PRIu32, expectedNbAreas
 				);
+			}
 			std::unique_ptr<Section> curSection = std::make_unique<Section>();
 
+			curSection->src = &where;
+			curSection->lineNo = lineNo;
+
 			getToken(line.data(), "'A' line is too short");
 			assume(strlen(token) != 0); // This should be impossible, tokens are non-empty
 			// The following is required for fragment offsets to be reliably predicted
 			for (FileSection &entry : fileSections) {
-				if (!strcmp(token, entry.section->name.c_str()))
+				if (!strcmp(token, entry.section->name.c_str())) {
 					fatal(&where, lineNo, "Area \"%s\" already defined earlier", token);
+				}
 			}
 			char const *sectName = token; // We'll deal with the section's name depending on type
 
@@ -269,21 +289,23 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 
 			uint32_t tmp = parseNumber(where, lineNo, token, numberType);
 
-			if (tmp > UINT16_MAX)
+			if (tmp > UINT16_MAX) {
 				fatal(
 				    &where,
 				    lineNo,
 				    "Area \"%s\" is larger than the GB address space!?",
 				    curSection->name.c_str()
 				);
+			}
 			curSection->size = tmp;
 
 			expectToken("flags", 'A');
 
 			getToken(nullptr, "'A' line is too short");
 			tmp = parseNumber(where, lineNo, token, numberType);
-			if (tmp & (1 << AREA_PAGING))
+			if (tmp & (1 << AREA_PAGING)) {
 				fatal(&where, lineNo, "Internal error: paging is not supported");
+			}
 			curSection->isAddressFixed = tmp & (1 << AREA_ISABS);
 			curSection->isBankFixed = curSection->isAddressFixed;
 			curSection->modifier = curSection->isAddressFixed || (tmp & (1 << AREA_TYPE))
@@ -341,17 +363,18 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 		}
 
 		case 'S': {
-			if (fileSymbols.size() == expectedNbSymbols)
-				warning(
+			if (fileSymbols.size() == expectedNbSymbols) {
+				error(
 				    &where,
 				    lineNo,
 				    "Got more 'S' lines than the expected %" PRIu32,
 				    expectedNbSymbols
 				);
+				break; // Refuse processing the line further.
+			}
 			Symbol &symbol = fileSymbols.emplace_back();
 
 			// Init other members
-			symbol.objFileName = where.name().c_str();
 			symbol.src = &where;
 			symbol.lineNo = lineNo;
 
@@ -378,6 +401,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 			// Expected format: /[DR]ef[0-9A-F]+/i
 			if (token[0] == 'R' || token[0] == 'r') {
 				symbol.type = SYMTYPE_IMPORT;
+				sym_AddSymbol(symbol);
 				// TODO: hard error if the rest is not zero
 			} else if (token[0] != 'D' && token[0] != 'd') {
 				fatal(&where, lineNo, "'S' line is neither \"Def\" nor \"Ref\"");
@@ -390,10 +414,11 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					// The same symbol can only be defined twice if neither
 					// definition is in a floating section
 					auto checkSymbol = [](Symbol const &sym) -> std::tuple<Section *, int32_t> {
-						if (auto *label = std::get_if<Label>(&sym.data); label)
-							return {label->section, label->offset};
-						assume(std::holds_alternative<int32_t>(sym.data));
-						return {nullptr, std::get<int32_t>(sym.data)};
+						if (sym.data.holds<Label>()) {
+							Label const &label = sym.data.get<Label>();
+							return {label.section, label.offset};
+						}
+						return {nullptr, sym.data.get<int32_t>()};
 					};
 					auto [symbolSection, symbolValue] = checkSymbol(symbol);
 					auto [otherSection, otherValue] = checkSymbol(*other);
@@ -402,16 +427,17 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					    || (symbolSection && !symbolSection->isAddressFixed)) {
 						sym_AddSymbol(symbol); // This will error out
 					} else if (otherValue != symbolValue) {
-						error(
-						    &where,
-						    lineNo,
-						    "Definition of \"%s\" conflicts with definition in %s (%" PRId32
-						    " != %" PRId32 ")",
+						fprintf(
+						    stderr,
+						    "error: \"%s\" is defined as %" PRId32 " at ",
 						    symbol.name.c_str(),
-						    other->objFileName,
-						    symbolValue,
-						    otherValue
+						    symbolValue
 						);
+						symbol.src->dump(symbol.lineNo);
+						fprintf(stderr, ", but as %" PRId32 " at ", otherValue);
+						other->src->dump(other->lineNo);
+						putc('\n', stderr);
+						exit(1);
 					}
 				} else {
 					// Add a new definition
@@ -420,11 +446,13 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 				// It's fine to keep modifying the symbol after `AddSymbol`, only
 				// the name must not be modified
 			}
-			if (strncasecmp(&token[1], "ef", 2) != 0)
+			if (strncasecmp(&token[1], "ef", 2) != 0) {
 				fatal(&where, lineNo, "'S' line is neither \"Def\" nor \"Ref\"");
+			}
 
-			if (!fileSections.empty())
+			if (!fileSections.empty()) {
 				fileSections.back().section->symbols.push_back(&symbol);
+			}
 
 			expectEol("'S' line is too long");
 			break;
@@ -432,15 +460,18 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 
 		case 'T':
 			// Now, time to parse the data!
-			if (!data.empty())
+			if (!data.empty()) {
 				warning(&where, lineNo, "Previous 'T' line had no 'R' line (ignored)");
+			}
 
 			data.clear();
-			for (token = strtok(line.data(), delim); token; token = strtok(nullptr, delim))
+			for (token = strtok(line.data(), delim); token; token = strtok(nullptr, delim)) {
 				data.push_back(parseByte(where, lineNo, token, numberType));
+			}
 
-			if (data.size() < ADDR_SIZE)
+			if (data.size() < ADDR_SIZE) {
 				fatal(&where, lineNo, "'T' line is too short");
+			}
 			// Importantly, now we know that there is "pending data" in `data`
 			break;
 
@@ -459,8 +490,8 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 			getToken(nullptr, "'R' line is too short");
 			areaIdx = parseByte(where, lineNo, token, numberType);
 			getToken(nullptr, "'R' line is too short");
-			areaIdx |= (uint16_t)parseByte(where, lineNo, token, numberType) << 8;
-			if (areaIdx >= fileSections.size())
+			areaIdx |= static_cast<uint16_t>(parseByte(where, lineNo, token, numberType)) << 8;
+			if (areaIdx >= fileSections.size()) {
 				fatal(
 				    &where,
 				    lineNo,
@@ -468,6 +499,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 				    areaIdx,
 				    fileSections.size()
 				);
+			}
 			assume(!fileSections.empty()); // There should be at least one, from the above check
 			Section *section = fileSections[areaIdx].section.get();
 			uint16_t *writeIndex = &fileSections[areaIdx].writeIndex;
@@ -475,7 +507,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 			uint16_t addr = data[0] | data[1] << 8;
 
 			if (section->isAddressFixed) {
-				if (addr < section->org)
+				if (addr < section->org) {
 					fatal(
 					    &where,
 					    lineNo,
@@ -485,12 +517,13 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					    section->name.c_str(),
 					    section->org
 					);
+				}
 				addr -= section->org;
 			}
 			// Lines are emitted that violate this check but contain no "payload";
 			// ignore those. "Empty" lines shouldn't trigger allocation, either.
 			if (data.size() != ADDR_SIZE) {
-				if (addr != *writeIndex)
+				if (addr != *writeIndex) {
 					fatal(
 					    &where,
 					    lineNo,
@@ -499,6 +532,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					    addr,
 					    *writeIndex
 					);
+				}
 				if (section->data.empty()) {
 					assume(section->size != 0);
 					section->data.resize(section->size);
@@ -523,14 +557,15 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 
 				if ((flags & 0xF0) == 0xF0) {
 					getToken(nullptr, "Incomplete relocation");
-					flags =
-					    (flags & 0x0F) | (uint16_t)parseByte(where, lineNo, token, numberType) << 4;
+					flags = (flags & 0x0F)
+					        | static_cast<uint16_t>(parseByte(where, lineNo, token, numberType))
+					              << 4;
 				}
 
 				getToken(nullptr, "Incomplete relocation");
 				uint8_t offset = parseByte(where, lineNo, token, numberType);
 
-				if (offset < ADDR_SIZE)
+				if (offset < ADDR_SIZE) {
 					fatal(
 					    &where,
 					    lineNo,
@@ -538,7 +573,8 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					    offset,
 					    ADDR_SIZE
 					);
-				if (offset >= data.size())
+				}
+				if (offset >= data.size()) {
 					fatal(
 					    &where,
 					    lineNo,
@@ -546,18 +582,21 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					    offset,
 					    data.size()
 					);
+				}
 
 				getToken(nullptr, "Incomplete relocation");
 				uint16_t idx = parseByte(where, lineNo, token, numberType);
 
 				getToken(nullptr, "Incomplete relocation");
-				idx |= (uint16_t)parseByte(where, lineNo, token, numberType);
+				idx |= static_cast<uint16_t>(parseByte(where, lineNo, token, numberType));
 
 				// Loudly fail on unknown flags
-				if (flags & (1 << RELOC_ZPAGE | 1 << RELOC_NPAGE))
+				if (flags & (1 << RELOC_ZPAGE | 1 << RELOC_NPAGE)) {
 					fatal(&where, lineNo, "Paging flags are not supported");
-				if (flags & ~RELOC_ALL_FLAGS)
+				}
+				if (flags & ~RELOC_ALL_FLAGS) {
 					warning(&where, lineNo, "Unknown reloc flags 0x%x", flags & ~RELOC_ALL_FLAGS);
+				}
 
 				// Turn this into a Patch
 				Patch &patch = section->patches.emplace_back();
@@ -567,7 +606,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 				patch.offset = offset - writtenOfs + *writeIndex;
 				if (section->patches.size() > 1) {
 					uint32_t prevOffset = section->patches[section->patches.size() - 2].offset;
-					if (prevOffset >= patch.offset)
+					if (prevOffset >= patch.offset) {
 						fatal(
 						    &where,
 						    lineNo,
@@ -576,6 +615,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 						    prevOffset,
 						    patch.offset
 						);
+					}
 				}
 				patch.pcSection = section;         // No need to fill `pcSectionID`, then
 				patch.pcOffset = patch.offset - 1; // For `jr`s
@@ -585,7 +625,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 				uint32_t baseValue = 0;
 
 				assume(offset < data.size());
-				if (data.size() - offset < nbBaseBytes)
+				if (data.size() - offset < nbBaseBytes) {
 					fatal(
 					    &where,
 					    lineNo,
@@ -593,13 +633,15 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					    nbBaseBytes,
 					    data.size() - offset
 					);
-				for (uint8_t i = 0; i < nbBaseBytes; ++i)
+				}
+				for (uint8_t i = 0; i < nbBaseBytes; ++i) {
 					baseValue = baseValue | data[offset + i] << (8 * i);
+				}
 
 				// Bit 4 specifies signedness, but I don't think that matters?
 				// Generate a RPN expression from the info and flags
 				if (flags & 1 << RELOC_ISSYM) {
-					if (idx >= fileSymbols.size())
+					if (idx >= fileSymbols.size()) {
 						fatal(
 						    &where,
 						    lineNo,
@@ -607,6 +649,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 						    idx,
 						    fileSymbols.size()
 						);
+					}
 					Symbol const &sym = fileSymbols[idx];
 
 					// SDCC has a bunch of "magic symbols" that start with a
@@ -616,10 +659,11 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 						// Look for the symbol being referenced, and use its index instead
 						for (idx = 0; idx < fileSymbols.size(); ++idx) {
 							if (sym.name.ends_with(fileSymbols[idx].name)
-							    && 1 + sym.name.length() == fileSymbols[idx].name.length())
+							    && 1 + sym.name.length() == fileSymbols[idx].name.length()) {
 								break;
+							}
 						}
-						if (idx == fileSymbols.size())
+						if (idx == fileSymbols.size()) {
 							fatal(
 							    &where,
 							    lineNo,
@@ -627,6 +671,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 							    sym.name.c_str(),
 							    &sym.name.c_str()[1]
 							);
+						}
 						patch.rpnExpression.resize(5);
 						patch.rpnExpression[0] = RPN_BANK_SYM;
 						patch.rpnExpression[1] = idx;
@@ -637,7 +682,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 						patch.rpnExpression.resize(1 + sym.name.length() - 2 + 1);
 						patch.rpnExpression[0] = RPN_SIZEOF_SECT;
 						memcpy(
-						    (char *)&patch.rpnExpression[1],
+						    reinterpret_cast<char *>(&patch.rpnExpression[1]),
 						    &sym.name.c_str()[2],
 						    sym.name.length() - 2 + 1
 						);
@@ -645,7 +690,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 						patch.rpnExpression.resize(1 + sym.name.length() - 2 + 1);
 						patch.rpnExpression[0] = RPN_STARTOF_SECT;
 						memcpy(
-						    (char *)&patch.rpnExpression[1],
+						    reinterpret_cast<char *>(&patch.rpnExpression[1]),
 						    &sym.name.c_str()[2],
 						    sym.name.length() - 2 + 1
 						);
@@ -658,7 +703,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 						patch.rpnExpression[4] = idx >> 24;
 					}
 				} else {
-					if (idx >= fileSections.size())
+					if (idx >= fileSections.size()) {
 						fatal(
 						    &where,
 						    lineNo,
@@ -666,6 +711,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 						    idx,
 						    fileSections.size()
 						);
+					}
 					// It gets funky. If the area is absolute, *actually*, we
 					// must not add its base address, as the assembler will
 					// already have added it in `baseValue`.
@@ -673,8 +719,9 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					// address from `baseValue`, undoing what the assembler did;
 					// this allows the relocation to still be correct, even if
 					// the section gets moved for any reason.
-					if (fileSections[idx].section->isAddressFixed)
+					if (fileSections[idx].section->isAddressFixed) {
 						baseValue -= fileSections[idx].section->org;
+					}
 					std::string const &name = fileSections[idx].section->name;
 					Section const *other = sect_GetSection(name);
 
@@ -686,12 +733,17 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					// so this fragment will be appended to the existing section
 					// *if any*, and thus its offset will be the section's
 					// current size.
-					if (other)
+					if (other) {
 						baseValue += other->size;
+					}
 					patch.rpnExpression.resize(1 + name.length() + 1);
 					patch.rpnExpression[0] = RPN_STARTOF_SECT;
 					// The cast is fine, it's just different signedness
-					memcpy((char *)&patch.rpnExpression[1], name.c_str(), name.length() + 1);
+					memcpy(
+					    reinterpret_cast<char *>(&patch.rpnExpression[1]),
+					    name.c_str(),
+					    name.length() + 1
+					);
 				}
 
 				patch.rpnExpression.push_back(RPN_CONST);
@@ -705,7 +757,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					// Despite the flag's name, as soon as it is set, 3 bytes
 					// are present, so we must skip two of them
 					if (flags & 1 << RELOC_EXPR16) {
-						if (*writeIndex + (offset - writtenOfs) > section->size)
+						if (*writeIndex + (offset - writtenOfs) > section->size) {
 							fatal(
 							    &where,
 							    lineNo,
@@ -714,6 +766,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 							    *writeIndex + (offset - writtenOfs),
 							    section->size
 							);
+						}
 						// Copy all bytes up to those (plus the byte that we'll overwrite)
 						memcpy(
 						    &section->data[*writeIndex], &data[writtenOfs], offset - writtenOfs + 1
@@ -770,7 +823,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 			// If there is some data left to append, do so
 			if (writtenOfs != data.size()) {
 				assume(data.size() > writtenOfs);
-				if (*writeIndex + (data.size() - writtenOfs) > section->size)
+				if (*writeIndex + (data.size() - writtenOfs) > section->size) {
 					fatal(
 					    &where,
 					    lineNo,
@@ -779,6 +832,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 					    *writeIndex + (data.size() - writtenOfs),
 					    section->size
 					);
+				}
 				memcpy(&section->data[*writeIndex], &data[writtenOfs], data.size() - writtenOfs);
 				*writeIndex += data.size() - writtenOfs;
 			}
@@ -798,9 +852,10 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 #undef expectToken
 #undef getToken
 
-	if (!data.empty())
+	if (!data.empty()) {
 		warning(&where, lineNo, "Last 'T' line had no 'R' line (ignored)");
-	if (fileSections.size() < expectedNbAreas)
+	}
+	if (fileSections.size() < expectedNbAreas) {
 		warning(
 		    &where,
 		    lineNo,
@@ -808,7 +863,8 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 		    expectedNbAreas,
 		    fileSections.size()
 		);
-	if (fileSymbols.size() < expectedNbSymbols)
+	}
+	if (fileSymbols.size() < expectedNbSymbols) {
 		warning(
 		    &where,
 		    lineNo,
@@ -816,6 +872,7 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 		    expectedNbSymbols,
 		    fileSymbols.size()
 		);
+	}
 
 	nbSectionsToAssign += fileSections.size();
 
@@ -823,7 +880,37 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 		std::unique_ptr<Section> &section = entry.section;
 
 		// RAM sections can have a size, but don't get any data (they shouldn't have any)
-		if (entry.writeIndex != section->size && entry.writeIndex != 0)
+		if (section->type != SECTTYPE_INVALID) {
+			auto const &typeInfo = sectionTypeInfo[section->type];
+			// Otherwise, how would the type already be known at this point?
+			assume(section->isAddressFixed);
+
+			if (!sect_HasData(section->type)) {
+				if (!section->data.empty()) {
+					fatal(
+					    &where,
+					    lineNo,
+					    "\"%s\" is implicitly defined as a %s section (being at address $%04" PRIx16
+					    "), but it has data! (Was a bad `__at()` value used?)",
+					    section->name.c_str(),
+					    typeInfo.name.c_str(),
+					    section->org
+					);
+				}
+			} else if (section->size != 0 && section->data.empty()) {
+				fatal(
+				    &where,
+				    lineNo,
+				    "\"%s\" is implicitly defined as a %s section (being at address $%04" PRIx16
+				    "), but it doesn't have any data! (Was a bad `__at()` value used?)",
+				    section->name.c_str(),
+				    typeInfo.name.c_str(),
+				    section->org
+				);
+			}
+		}
+
+		if (entry.writeIndex != 0 && entry.writeIndex != section->size) {
 			fatal(
 			    &where,
 			    lineNo,
@@ -832,14 +919,26 @@ void sdobj_ReadFile(FileStackNode const &where, FILE *file, std::vector<Symbol>
 			    entry.writeIndex,
 			    section->size
 			);
-
-		if (section->modifier == SECTION_FRAGMENT) {
-			// Add the fragment's offset to all of its symbols
-			for (Symbol *symbol : section->symbols)
-				symbol->label().offset += section->offset;
 		}
-
 		// Calling `sect_AddSection` invalidates the contents of `fileSections`!
 		sect_AddSection(std::move(section));
 	}
+
+	// Fix symbols' section pointers to component sections
+	// This has to run **after** all the `sect_AddSection()` calls,
+	// so that `sect_GetSection()` will work
+	for (Symbol &sym : fileSymbols) {
+		if (sym.data.holds<Label>()) {
+			Label &label = sym.data.get<Label>();
+			if (Section *section = label.section; section->modifier != SECTION_NORMAL) {
+				if (section->modifier == SECTION_FRAGMENT) {
+					// Add the fragment's offset to the symbol's
+					// (`section->offset` is computed by `sect_AddSection`)
+					label.offset += section->offset;
+				}
+				// Associate the symbol with the main section, not the "component" one
+				label.section = sect_GetSection(section->name);
+			}
+		}
+	}
 }
